import collections.abc
import functools
import logging
import traceback

import requests

import ci.util
import ccc.elasticsearch
import ctx
import oci.auth as oa
import oci.client as oc
import model.container_registry


logger = logging.getLogger(__name__)


@functools.lru_cache
def oci_cfg_lookup(
    cfg_factory=None,
) -> collections.abc.Callable[[str, oa.Privileges, bool], oa.OciCredentials]:
    def find_credentials(
        image_reference: str,
        privileges: oa.Privileges=oa.Privileges.READONLY,
        absent_ok: bool=True,
    ):
        registry_cfg = model.container_registry.find_config(
            image_reference=image_reference,
            privileges=privileges,
            cfg_factory=cfg_factory,
        )
        if not registry_cfg:
            if absent_ok:
                return None # fallback to docker-cfg (or try w/o auth)
            else:
                raise RuntimeError(
                f'No credentials found for {image_reference} with {privileges=}'
            )
        creds = registry_cfg.credentials()
        return oa.OciBasicAuthCredentials(
            username=creds.username(),
            password=creds.passwd(),
        )

    return find_credentials


def oci_request_handler_requirements_fulfilled() -> bool:
    '''
    checks requirements for oci request handler installation
    returns False if a requirement is not fulfilled and prints a warning with reason
    '''
    try:
        cfg_set = ctx.cfg_set()
        cfg_set.elasticsearch()
        return True
    except ValueError:
        logger.warning('no elasticsearch config found')
        return False


@functools.cache
def oci_client(
    credentials_lookup: collections.abc.Callable=None,
    install_logging_handler: bool=True,
    cfg_factory=None,
    http_connection_pool_size:int=16,
    tag_preprocessing_callback: collections.abc.Callable[[str], str]=None,
    tag_postprocessing_callback: collections.abc.Callable[[str], str]=None,
) -> oc.Client:
    def base_api_lookup(image_reference):
        registry_cfg = model.container_registry.find_config(
            image_reference=image_reference,
            privileges=None,
            cfg_factory=cfg_factory,
        )
        if registry_cfg and (base_url := registry_cfg.api_base_url()):
            return base_url
        return oc.base_api_url(image_reference)

    routes = oc.OciRoutes(base_api_lookup)

    install_logging_handler &= ci.util._running_on_ci()
    if install_logging_handler:
        try:
            if oci_request_handler_requirements_fulfilled():
                _add_oci_request_logging_handler_unless_already_registered()
            else:
                logger.warning('skipping oci request logger installation')
        except:
            # do not fail just because of logging-issue
            import traceback
            traceback.print_exc()

    if not credentials_lookup:
        credentials_lookup = oci_cfg_lookup(cfg_factory=cfg_factory)

    # increase poolsize (defaults: 10) to allow for greater parallelism
    session = requests.Session()
    adapter = requests.adapters.HTTPAdapter(
        pool_connections=http_connection_pool_size,
        pool_maxsize=http_connection_pool_size,
    )
    session.mount('https://', adapter)

    return oc.Client(
        credentials_lookup=credentials_lookup,
        routes=routes,
        session=session,
        tag_preprocessing_callback=tag_preprocessing_callback,
        tag_postprocessing_callback=tag_postprocessing_callback,
    )


class _OciRequestHandler(logging.Handler):
    def __init__(
        self,
        level,
        es_client,
        *args,
        **kwargs,
    ) -> None:
        self.es_client = es_client
        self.shortcut = False
        super().__init__(level=level, *args, **kwargs)

    def emit(self, record: logging.LogRecord) -> None:
        if self.shortcut:
            logger.info('oci-request-reporting to elasticsearch was shortcut due to previous error')
            return

        method = record.__dict__.get('method')
        url = record.__dict__.get('url')
        try:
            self.es_client.store_document(
                index='oci_request',
                body={
                    'method': method,
                    'url': url,
                    'stacktrace': traceback.format_stack(),
                }
            )
        except:
            logger.warning(traceback.format_exc())
            logger.warning('could not send oci request log to elastic search')
            self.shortcut = True


_client_sentinel = object()
_es_client = _client_sentinel # init by first caller
_es_handler = None


def _add_oci_request_logging_handler_unless_already_registered():
    global _es_client
    global _es_handler

    if _es_client is _client_sentinel:
        _es_client = ccc.elasticsearch.default_client_if_available()

    if not _es_client:
        return

    if not _es_handler:
        _es_handler = _OciRequestHandler(level=logging.DEBUG, es_client=_es_client)

    es_logger = logging.getLogger('oci.client.request_logger')
    es_logger.setLevel(logging.DEBUG)

    if not _es_handler in es_logger.handlers:
        es_logger.addHandler(_es_handler)



import collections.abc
import dataclasses
import enum
import functools
import json
import typing
import urllib.parse

import dacite

import oci.util

OCI_MANIFEST_SCHEMA_V2_MIME = 'application/vnd.oci.image.manifest.v1+json'
OCI_IMAGE_INDEX_MIME = 'application/vnd.oci.image.index.v1+json'

DOCKER_MANIFEST_LIST_MIME = 'application/vnd.docker.distribution.manifest.list.v2+json'
DOCKER_MANIFEST_SCHEMA_V2_MIME = 'application/vnd.docker.distribution.manifest.v2+json'

empty_dict = dataclasses.field(default_factory=dict) # noqa:E3701


class MimeTypes:
    '''
    predefined, well-known mimetypes, handy to be used in oci.client.Client.manifest as `accept` arg

    single_image: force single-image
    multiarch: force multi-arch (image-list)
    prefer_multiarch

    note: not all registries honour `access` header
    '''
    single_image = ', '.join((OCI_MANIFEST_SCHEMA_V2_MIME, DOCKER_MANIFEST_SCHEMA_V2_MIME))
    multiarch = ', '.join((OCI_IMAGE_INDEX_MIME, DOCKER_MANIFEST_LIST_MIME))
    prefer_multiarch = ', '.join((multiarch, single_image))


class OciTagType(enum.Enum):
    SYMBOLIC = 'symbolic'
    DIGEST = 'digest'
    NO_TAG = 'no_tag'


class OciRegistryType(enum.Enum):
    GAR = 'gar'
    GCR = 'gcr'
    DOCKERHUB = 'dockerhub'
    ARTIFACTORY = 'artifactory'
    GHCR = 'ghcr'
    ALIYUN = 'aliyun'
    UNKNOWN = 'unknown'

    @staticmethod
    def from_image_ref(
        image_reference: typing.Union[str, 'OciImageReference'],
    ) -> 'OciRegistryType':
        netloc = OciImageReference.to_image_ref(image_reference=image_reference).netloc

        if 'pkg.dev' in netloc:
            return OciRegistryType.GAR
        if 'gcr.io' in netloc:
            return OciRegistryType.GCR
        if 'docker.io' in netloc:
            return OciRegistryType.DOCKERHUB
        if 'ghcr.io' in netloc:
            return OciRegistryType.GHCR
        if 'cr.aliyuncs.com' in netloc:
            return OciRegistryType.ALIYUN

        return OciRegistryType.UNKNOWN


class OciImageReference:
    @staticmethod
    def to_image_ref(
        image_reference: typing.Union[str, 'OciImageReference'],
        normalise: bool=True,
    ):
        if isinstance(image_reference, OciImageReference):
            return image_reference
        else:
            return OciImageReference(
                image_reference=image_reference,
                normalise=normalise,
            )

    def __init__(
        self,
        image_reference: typing.Union[str, 'OciImageReference'],
        normalise: bool=True,
    ):
        if isinstance(image_reference, OciImageReference):
            self._orig_image_reference = image_reference._orig_image_reference
        elif isinstance(image_reference, str):
            self._orig_image_reference = image_reference
        else:
            raise ValueError(image_reference)
        self._normalise = normalise

    @property
    def original_image_reference(self) -> str:
        return self._orig_image_reference

    @property
    @functools.cache
    def normalised_image_reference(self) -> str:
        return oci.util.normalise_image_reference(self._orig_image_reference)

    @property
    @functools.cache
    def netloc(self) -> str:
        return self.urlparsed.netloc

    @property
    @functools.cache
    def ref_without_tag(self) -> str:
        '''
        returns the (normalised) image reference w/o the tag or digest tag.
        '''
        p = self.urlparsed
        name = p.netloc + p.path.rsplit('@', 1)[0].rsplit(':', 1)[0]

        return name

    @property
    @functools.cache
    def name(self) -> str:
        '''
        returns the (normalised) image name (omitting api-prefix and tag)
        '''
        p = self.urlparsed
        name = p.path[1:].rsplit('@', 1)[0].rsplit(':', 1)[0]

        return name

    @property
    @functools.cache
    def has_digest_tag(self) -> bool:
        return self.tag_type is OciTagType.DIGEST

    @property
    @functools.cache
    def has_symbolical_tag(self) -> bool:
        return self.tag_type is OciTagType.SYMBOLIC

    @property
    @functools.cache
    def has_mixed_tag(self) -> bool:
        if not self.has_digest_tag:
            return False

        p = self.urlparsed
        ref_without_digest_tag = p.netloc + p.path.rsplit('@', 1)[0]

        return ':' in ref_without_digest_tag

    @property
    @functools.cache
    def with_symbolical_tag(self) -> str:
        if not (self.has_symbolical_tag or self.has_mixed_tag):
            raise ValueError(f'does not contain a symbolical tag: {str(self)=}')

        p = self.urlparsed
        return p.netloc + p.path.rsplit('@', 1)[0]

    @property
    @functools.cache
    def has_tag(self):
        return not self.tag_type is OciTagType.NO_TAG

    @property
    @functools.cache
    def tag(self) -> str:
        p = self.urlparsed

        if '@' in p.path:
            return p.path.rsplit('@', 1)[-1]
        elif ':' in p.path:
            return p.path.rsplit(':', 1)[-1]
        else:
            raise ValueError(f'no tag found for {str(self)}')

    @property
    @functools.cache
    def tag_type(self) -> OciTagType:
        p = self.urlparsed

        if '@' in p.path:
            return OciTagType.DIGEST
        elif ':' in p.path:
            return OciTagType.SYMBOLIC
        else:
            return OciTagType.NO_TAG

    @property
    def parsed_mixed_tag(self) -> tuple[str, str]:
        if not self.has_mixed_tag:
            raise ValueError(f'not a mixed-tag: {str(self)=}')

        p = self.urlparsed
        digest_tag = p.path.rsplit('@', 1)[-1]
        symbolical_tag = p.path.rsplit('@', 1)[0].rsplit(':', 1)[-1]
        return symbolical_tag, digest_tag

    @property
    def parsed_digest_tag(self) -> tuple[str, str]:
        if not self.tag_type is OciTagType.DIGEST:
            raise ValueError(f'not a digest-tag: {str(self)=}')

        algorithm, digest = self.tag.split(':')
        return algorithm, digest

    @property
    def digest(self) -> str:
        _, digest = self.parsed_digest_tag
        return digest

    @property
    def local_ref(self) -> str:
        return self.urlparsed.path.removeprefix('/')

    @property
    @functools.cache
    def urlparsed(self) -> urllib.parse.ParseResult:
        if not '://' in (img_ref := str(self)) and not img_ref.startswith('/'):
            return urllib.parse.urlparse(f'https://{img_ref}')
        return urllib.parse.urlparse(img_ref)

    def with_tag(self, tag: str) -> str:
        if 'sha256' in tag and not '@' in tag:
            return f'{self.ref_without_tag}@{tag}'
        return f'{self.ref_without_tag}:{tag}'

    def __str__(self) -> str:
        if self._normalise:
            return self.normalised_image_reference
        return self._orig_image_reference

    def __repr__(self) -> str:
        return f'OciImageReference({str(self)})'

    def __eq__(self, other) -> bool:
        if not isinstance(other, OciImageReference):
            # XXX: should we return True for str with same value?
            return False

        if self._orig_image_reference == other._orig_image_reference:
            return True

        return oci.util.normalise_image_reference(self._orig_image_reference) == \
               oci.util.normalise_image_reference(other._orig_image_reference)

    def __hash__(self):
        return hash((self._orig_image_reference,))


class OciManifestSchemaVersion(enum.Enum):
    V1 = 1
    V2 = 2


class OciImageNotFoundException(Exception):
    pass


@dataclasses.dataclass(frozen=True)
class OciBlobRef:
    digest: str
    mediaType: str
    size: int
    annotations: dict | None = None

    def as_dict(self) -> dict:
        raw = dataclasses.asdict(self)
        # fields that are None should not be included in the output
        raw = {k:v for k,v in raw.items() if v is not None}
        return raw

    def __hash__(self):
        annotations = tuple(sorted(self.annotations.items())) if self.annotations else ()
        return hash((self.digest, self.size, self.mediaType, annotations))

    def __eq__(self, other) -> bool:
        if not isinstance(other, OciBlobRef):
            return False

        other: OciBlobRef

        if not self.digest == other.digest:
            return False

        if not self.mediaType == other.mediaType:
            return False

        if not self.size == other.size:
            return False

        if not self.annotations == other.annotations:
            return False

        return True


@dataclasses.dataclass
class OciImageManifest:
    config: OciBlobRef
    layers: collections.abc.Sequence[OciBlobRef]
    mediaType: str = OCI_MANIFEST_SCHEMA_V2_MIME
    schemaVersion: int = 2
    annotations: dict = dataclasses.field(default_factory=dict)

    def as_dict(self) -> dict:
        def layer_to_dict(layer):
            if isinstance(layer, dict):
                return layer
            else:
                return layer.as_dict()

        return {
            'config': self.config.as_dict(),
            'layers': [layer_to_dict(layer) for layer in self.layers],
            'mediaType': self.mediaType,
            'schemaVersion': self.schemaVersion,
            'annotations': self.annotations,
        }

    def blobs(self) -> collections.abc.Generator[OciBlobRef, None, None]:
        yield self.config
        yield from self.layers


@dataclasses.dataclass
class OciBlobRefV1:
    blobSum: str


@dataclasses.dataclass
class OciImageManifestV1:
    '''
    defines replication-relevant parts of the (deprecated) oci-manifest-schema version 1

    note: the `layers` attr must be initialised after deserialisation with v2-compatible
    objects of `OciBlobRef` (see oci/client.py for reference)

    this class is not intended to be instantiated by users of this module
    '''
    name: str
    tag: str
    architecture: str
    fsLayers: list[OciBlobRefV1]
    history: list[dict] # don't care about details
    signatures: list[dict] = empty_dict # don't care about details
    schemaVersion: int = 1
    layers = None # to be initialised by factory-function

    def blobs(self) -> collections.abc.Generator[OciBlobRef, None, None]:
        if not self.layers:
            raise ValueError('instance was not properly initialised')

        yield from self.layers # noqa


@dataclasses.dataclass(frozen=True)
class OciPlatform:
    '''
    https://github.com/distribution/distribution/blob/main/docs/spec/manifest-v2-2.md#manifest-list
    '''
    architecture: str
    os: str # could also be a dict (see spec)
    variant: str | None = None
    features: list[str] | None = None

    def as_dict(self) -> dict:
        # need custom serialisation, because some OCI registries do not like null-values
        # (must be absent instead)
        raw = dataclasses.asdict(self)

        if not self.variant:
            del raw['variant']

        if not self.features:
            del raw['features']

        return raw

    def __eq__(self, other):
        if not isinstance(other, OciPlatform):
            return False

        if not self.architecture == other.architecture:
            return False

        if not self.os == other.os:
            return False

        return True


@dataclasses.dataclass(frozen=True)
class OciImageManifestListEntry(OciBlobRef):
    artifactType: str | None = None
    data: str | None = None
    platform: OciPlatform | None = None
    urls: list[str] | None = None

    def as_dict(self) -> dict:
        raw = OciBlobRef.as_dict(self)
        # platform is an optional attribute according to oci spec
        # => only include in the output if set
        if self.platform:
            raw['platform'] = self.platform.as_dict()
        # fields that are None should not be included in the output
        raw = {k:v for k,v in raw.items() if v is not None}
        return raw


@dataclasses.dataclass
class OciImageManifestList:
    '''Covers both Docker Manifest List
        (https://github.com/distribution/distribution/blob/main/docs/spec/manifest-v2-2.md#manifest-list)
        and OCI Image Index
        (https://github.com/opencontainers/image-spec/blob/main/image-index.md)
    '''
    manifests: list[OciImageManifestListEntry]
    mediaType: str = OCI_IMAGE_INDEX_MIME
    schemaVersion: int = 2
    annotations: dict = dataclasses.field(default_factory=dict)

    def as_dict(self):
        raw = {
            'manifests': [le.as_dict() for le in self.manifests],
            'mediaType': self.mediaType,
            'schemaVersion': self.schemaVersion,
        }

        if self.mediaType == OCI_IMAGE_INDEX_MIME:
            raw['annotations'] = self.annotations

        return raw


def as_manifest(
    manifest: str | bytes | dict | OciImageManifest | OciImageManifestList,
) -> OciImageManifest | OciImageManifestList:
    '''
    returns a deserialised equivalent of the passed-in manifest. For convenience, if passed-in
    manifest is already an instance of either OciImageManifest or OciImageManifestList, the
    passed value is returned unchanged.
    '''
    if isinstance(manifest, (OciImageManifest, OciImageManifestList)):
        return manifest

    if isinstance(manifest, (str, bytes)):
        manifest = json.loads(manifest)

    if (media_type := manifest.get('mediaType')) in (
        DOCKER_MANIFEST_LIST_MIME,
        OCI_IMAGE_INDEX_MIME,
    ):
        return dacite.from_dict(
            data_class=OciImageManifestList,
            data=manifest,
        )

    if media_type in (
        DOCKER_MANIFEST_SCHEMA_V2_MIME,
        OCI_MANIFEST_SCHEMA_V2_MIME,
    ):
        return dacite.from_dict(
            data_class=OciImageManifest,
            data=manifest,
        )

    raise ValueError(manifest, 'unknown schema-version')



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import (
    NamedModelElement,
)


class AlicloudConfig(NamedModelElement):
    def region(self):
        return self.raw.get('region')

    def access_key_id(self):
        return self.raw.get('access_key_id')

    def access_key_secret(self):
        return self.raw.get('access_key_secret')

    def _required_attributes(self):
        return ['region', 'access_key_id', 'access_key_secret']



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import ocm

import reutil


def component_ref_component_name_filter(include_regexes=(), exclude_regexes=()):
    if not include_regexes and not exclude_regexes:
        return lambda component: True

    def to_component_name(component: ocm.ComponentReference):
        return component.componentName

    return reutil.re_filter(
        include_regexes=include_regexes,
        exclude_regexes=exclude_regexes,
        value_transformation=to_component_name,
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import datetime

from concourse.model.job import JobVariant
from concourse.model.base import Trait, TraitTransformer, AttributeSpec


@dataclasses.dataclass
class TimeRange:
    begin: datetime.time
    end: datetime.time


ATTRIBUTES = (
    AttributeSpec.optional(
        name='interval',
        default='5m',
        doc='''
        go-style time interval between job executions. Supported suffixes are: `s`, `m`, `h`
        ''',
    ),
    AttributeSpec.optional(
        name='timezone',
        default='Europe/Berlin',
        doc='''
        timezone to use for start/end-times (if not configured, timezone has no effect)
        `List of timezones <https://en.wikipedia.org/wiki/List_of_tz_database_time_zones>`_
        ''',
    ),
    AttributeSpec.optional(
        name='days',
        default=None,
        type=list[str] | str,
        doc='''
        If set, specifies the weekdays on which to run the job. Has two forms:
        - list of strings (Monday, Tuesday, Wednesday, Thursday, Friday, Saturday, Sunday)
        - single string (WORKING_DAYS - i.e. MO..FR)
        ''',
    ),
    AttributeSpec.optional(
        name='timerange',
        default=None,
        type=str | dict,
        doc='''
        If set, specifies the time range in between which to run the job.

        Has two forms:

            - `WORKING_HOURS`, which limits the job to be triggered between 08:00 and 18:00.
            - A dict with `begin` and `end` in the format of `HH:MM`, e.g.:

                .. code-block:: yaml

                    timerange:
                        begin: '09:45'
                        end: '13:30'

            The timerange is interpreted relative to the configured timezone

        ''',
    ),
)


class CronTrait(Trait):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    @property
    def interval(self):
        return self.raw['interval']

    @property
    def timezone(self):
        return self.raw['timezone']

    @property
    def days(self):
        if not (days := self.raw.get('days')):
            return None

        work_days = (
            'Monday',
            'Tuesday',
            'Wednesday',
            'Thursday',
            'Friday',
        )
        weekend_days = (
            'Saturday',
            'Sunday',
        )
        all_days = work_days + weekend_days

        if isinstance(days, str):
            if days == 'WORKING_DAYS':
                return work_days
            else:
                raise ValueError('days must equal WORKING_DAYS if set to a single str')
        if isinstance(days, list):
            for d in days:
                if d not in all_days:
                    raise ValueError(f'{d} not in {all_days}')
            return days
        else:
            raise ValueError(days)

    @property
    def timerange(self):
        if not (timerange := self.raw['timerange']):
            return None

        if isinstance(timerange, str):
            if timerange == 'WORKING_HOURS':
                return TimeRange(
                    begin=datetime.time(hour=8, minute=0),
                    end=datetime.time(hour=20, minute=0),
                )
            else:
                raise ValueError(timerange)
        elif isinstance(timerange, dict):
            return TimeRange(
                begin=datetime.time.fromisoformat(timerange['begin']),
                end=datetime.time.fromisoformat(timerange['end']),
            )
        else:
            raise ValueError(timerange)

    def resource_name(self):
        # variant-names must be unique, so this should suffice, unless there are different
        # cron-resources (e.g. different days/timeranges)

        name = '-'.join((
            self.variant_name,
            self.interval,
        ))

        if tr := self.timerange:
            name += f'-{tr.begin.strftime("%H%M")}-{tr.end.strftime("%H%M")}'

        name += '-cron'

        return name

    def transformer(self):
        return CronTraitTransformer()


class CronTraitTransformer(TraitTransformer):
    name = 'cronjob'

    def process_pipeline_args(self, pipeline_args: JobVariant):
        main_repo = pipeline_args.main_repository()
        if main_repo:
            if 'trigger' not in pipeline_args.raw['repo']:
                main_repo._trigger = False
        # todo: inject cron-resource



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from contextlib import contextmanager
from io import StringIO
import sys
import typing


@contextmanager
def capture_out():
    new_stdout, new_stderr = StringIO(), StringIO()
    old_stdout, old_stderr = sys.stdout, sys.stderr
    try:
        sys.stdout, sys.stderr = new_stdout, new_stderr
        yield sys.stdout, sys.stderr
    finally:
        sys.stdout, sys.stderr = old_stdout, old_stderr


@contextmanager
def replace_modules(modules:dict):
    '''
    ctx manager that will replace the given modules in sys.modules and restore them
    afterwards
    '''
    original_modules = {
        name: sys.modules[name] for name in modules.keys()
    }
    try:
        for name, module in modules.items():
            sys.modules[name] = module
        yield None
    finally:
        for name, module in original_modules.items():
            sys.modules[name] = module


class AssertMixin:
    def assertEmpty(self, iterable, msg=None):
        if issubclass(type(iterable), typing.Sequence):
            leng = len(iterable)
            if leng == 0:
                return
            raise self.failureException('iterable was not empty')
        try:
            next(iterable)
            raise self.failureException('iterable was not empty')
        except StopIteration:
            return # ok - iterable was empty
        except Exception as e:
            raise self.failureException(str(e))



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import logging
import traceback

import falcon # pylint: disable=E0401

import ccc.elasticsearch
from .webhook import GithubWebhook
from model.webhook_dispatcher import WebhookDispatcherConfig
import whd.metric


logger = logging.getLogger(__name__)


def webhook_dispatcher_app(
    cfg_factory,
    cfg_set,
    whd_cfg: WebhookDispatcherConfig,
):
    es_client = ccc.elasticsearch.from_cfg(cfg_set.elasticsearch())
    if es_client:
        logger.info('will write webhook metrics to ES')

    def handle_exception(ex, req, resp, params):
        if not es_client:
            raise falcon.HTTPInternalServerError
        exc_trace = traceback.format_exc()
        logger.error(exc_trace)
        req_body = req.media
        exception_metric = whd.metric.ExceptionMetric.create(
            service='whd',
            stacktrace=exc_trace,
            request=req_body,
            params=params,
        )
        ccc.elasticsearch.metric_to_es(
            es_client=es_client,
            metric=exception_metric,
            index_name=whd.metric.index_name(exception_metric),
        )
        # raise HTTP error to not leak logs to client
        raise falcon.HTTPInternalServerError # noqa

    app = falcon.App(
        middleware=[],
    )

    app.add_route('/github-webhook',
        GithubWebhook(
            cfg_factory=cfg_factory,
            whd_cfg=whd_cfg,
            cfg_set=cfg_set,
            es_client=es_client,
        ),
    )
    app.add_error_handler(
        exception=Exception,
        handler=handle_exception,
    )

    return app



import os

own_dir = os.path.abspath(os.path.dirname(__file__))
repo_root_dir = os.path.join(own_dir, os.pardir)
res_dir = os.path.join(own_dir, 'resources')
template_dir = os.path.join(own_dir, 'templates')
template_include_dir = own_dir

last_released_tag_file = os.path.join(res_dir, 'LAST_RELEASED_TAG')

# available in cc-job-image, only
cc_bin_dir = '/cc/utils/bin'
launch_dockerd = os.path.join(cc_bin_dir, 'launch-dockerd.sh')



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import enum
import typing

from model.base import (
    ModelBase,
    NamedModelElement,
)


class ServiceAccountConfig(ModelBase):
    def _required_attributes(self):
        return {
            'name',
            'namespace',
        }

    def name(self) -> str:
        return self.raw['name']

    def namespace(self) -> str:
        return self.raw['namespace']


class RotationStrategy(enum.StrEnum):
    SECRET = 'secret'
    TOKEN_REQUEST = 'tokenRequest'


class KubernetesConfig(NamedModelElement):
    def _required_attributes(self):
        return {
            'kubeconfig',
        }

    def service_account(self) -> typing.Union[ServiceAccountConfig, None]:
        if raw_cfg := self.raw.get('service_account'):
            return ServiceAccountConfig(raw_dict=raw_cfg)
        else:
            return None

    def kubeconfig(self) -> dict:
        return self.raw.get('kubeconfig')

    def cluster_domain(self):
        return self.raw.get('cluster_domain')

    def namespace(self):
        '''
        fallback to default if no namespace is configured
        '''
        return self.raw.get('namespace', 'default')

    def rotation_strategy(self) -> RotationStrategy:
        raw = self.raw.get('rotation_strategy', 'secret')
        return RotationStrategy(raw)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import unittest

from concourse.client import routes


class ConcourseApiRoutesBaseTest(unittest.TestCase):
    def setUp(self):
        self.examinee = routes.ConcourseApiRoutesBase(
            base_url='https://made-up-concourse.com',
            team='foo'
        )

    def test_team_route(self):
        self.assertEqual(self.examinee.team_url(), 'https://made-up-concourse.com/api/v1/teams/foo')
        self.assertEqual(
            self.examinee.team_url(team='bar'),
            'https://made-up-concourse.com/api/v1/teams/bar'
        )

    def test_login_route(self):
        self.assertEqual(
            self.examinee.login(),
            'https://made-up-concourse.com/sky/token'
        )

    def test_pipelines_route(self):
        self.assertEqual(
            self.examinee.pipelines(),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines'
        )

    def test_order_pipelines_route(self):
        self.assertEqual(
            self.examinee.order_pipelines(),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines/ordering',
        )

    def test_pipeline_route(self):
        self.assertEqual(
            self.examinee.pipeline(pipeline_name='baz'),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines/baz',
        )

    def test_pipeline_config_route(self):
        self.assertEqual(
            self.examinee.pipeline_cfg(pipeline_name='baz'),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines/baz/config',
        )

    def test_unpause_pipeline_route(self):
        self.assertEqual(
            self.examinee.unpause_pipeline(pipeline_name='baz'),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines/baz/unpause',
        )

    def test_unpause_expose_route(self):
        self.assertEqual(
            self.examinee.expose_pipeline(pipeline_name='baz'),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines/baz/expose',
        )

    def test_resource_check_route(self):
        self.assertEqual(
            self.examinee.resource_check(pipeline_name='baz', resource_name='bar'),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines/baz/resources/bar/check',
        )

    def test_job_build_route(self):
        self.assertEqual(
            self.examinee.job_build(pipeline_name='baz', job_name='bar', build_name='123'),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines/baz/jobs/bar/builds/123',
        )

    def test_job_builds_route(self):
        self.assertEqual(
            self.examinee.job_builds(pipeline_name='baz', job_name='bar'),
            'https://made-up-concourse.com/api/v1/teams/foo/pipelines/baz/jobs/bar/builds',
        )

    def test_build_events_route(self):
        self.assertEqual(
            self.examinee.build_events(build_id=252525),
            'https://made-up-concourse.com/api/v1/builds/252525/events',
        )

    def test_build_plan_route(self):
        self.assertEqual(
            self.examinee.build_plan(build_id=252525),
            'https://made-up-concourse.com/api/v1/builds/252525/plan',
        )

    def test_list_worker_route(self):
        self.assertEqual(
            self.examinee.list_workers(),
            'https://made-up-concourse.com/api/v1/workers',
        )

    def test_prune_worker_route(self):
        self.assertEqual(
            self.examinee.prune_worker(worker_name='foo'),
            'https://made-up-concourse.com/api/v1/workers/foo/prune',
        )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from concourse.model.pipeline import PipelineDefinition
from concourse.model.job import JobVariant


class PipelineDefinitionValidator:
    def __init__(self, pipeline_definition):
        if not isinstance(pipeline_definition, PipelineDefinition):
            raise ValueError('not a PipelineDefinition: ' + str(pipeline_definition))

        self._pipeline_definition = pipeline_definition

    def validate(self):
        for variant in self._pipeline_definition.variants():
            self._validate_variant(variant)

    def _validate_variant(self, variant: JobVariant):
        if not isinstance(variant, JobVariant):
            raise ValueError('not a JobVariant: ' + str(variant))
        self._validate_element(variant)

    def _validate_element(self, element, parents=set()):
        for child in element._children():
            self._validate_element(child, parents | {element})
        element.validate()



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import functools
import pytest

import github.util as ghu

import ocm


# test gear
pull_request_mock = object() # keep this as simple as possible for now
create_upgrade_pr = functools.partial(
    ghu.UpgradePullRequest,
    pull_request=pull_request_mock,
)


def test_ctor():
    # upgrade component
    create_upgrade_pr(
        from_ref=ocm.ComponentReference(
            name='abcd', componentName='a.b/c1', version='1.2.3'
        ),
        to_ref=ocm.ComponentReference(
            name='abcd', componentName='a.b/c1', version='2.0.0'
        ),
    )
    # upgrade web dependency
    create_upgrade_pr(
        from_ref=ocm.Resource(
            name='dep_red',
            version='1.2.3',
            type=ocm.ArtefactType.BLOB,
            access=None,
        ),
        to_ref=ocm.Resource(
            name='dep_red',
            version='2.0.0',
            type=ocm.ArtefactType.BLOB,
            access=None,
        ),
    )
    # error: mismatch in dependency name
    with pytest.raises(ValueError, match='reference name mismatch'):
        create_upgrade_pr(
            from_ref=ocm.ComponentReference(
                name='foo', componentName='a.b/c1', version='1.2.3'
            ),
            to_ref=ocm.ComponentReference(
                name='bar', componentName='a.b/c1', version='2.0.0'
            ),
        )
    # error: mismatch in dependency types
    with pytest.raises(ValueError, match='reference types do not match'):
        create_upgrade_pr(
            from_ref=ocm.ComponentReference(
                name='dep_red', componentName='a.b/c1', version='1.2.3'
            ),
            to_ref=ocm.Resource(
                name='dep_red',
                version='2.0.0',
                type=ocm.ArtefactType.BLOB,
                access=None,
            ),
        )


def test_is_obsolete():
    examinee = create_upgrade_pr(
        from_ref=ocm.ComponentReference(
            name='c1',
            componentName='c1',
            version='1.2.3',
        ),
        to_ref=ocm.ComponentReference(
            name='c1',
            componentName='c1',
            version='2.0.0',
        ),
    )

    cref = ocm.ComponentReference(
        name='c1',
        componentName='c1',
        version='6.0.0',
    )

    reference_component = ocm.Component(
        name='c1',
        version='6.6.6',
        repositoryContexts=(),
        provider=None,
        sources=(),
        resources=(),
        componentReferences=()
    )

    # test with reference component not declaring this dependency
    assert not examinee.is_obsolete(reference_component=reference_component)

    # add differently-named dependency with greater version
    reference_component.componentReferences = (
        dataclasses.replace(cref, componentName='other-name'),
    )
    assert not examinee.is_obsolete(reference_component=reference_component)

    # add same-named web dependency with lesser version
    reference_component.componentReferences = (
        dataclasses.replace(cref, version='0.0.1'),
    )
    assert not examinee.is_obsolete(reference_component=reference_component)

    # add same-named resource of greater version but different type
    # todo: we should actually also test dependencies towards resources of two different types
    reference_component.resources = (
        ocm.Resource(
            name='c1',
            version='6.0.0',
            type=ocm.ArtefactType.BLOB,
            access=None,
        ),
    )
    assert not examinee.is_obsolete(reference_component=reference_component)

    # finally, add greater dependency of matching type and name
    reference_component.componentReferences = (
        dataclasses.replace(cref, version='9.9.9'),
    )
    assert examinee.is_obsolete(reference_component=reference_component)


def test_target_matches():
    old_resource = ocm.Resource(
        name='res1',
        version='1.2.3',
        type=ocm.ArtefactType.BLOB,
        access=ocm.Access(),
    )
    new_resource = ocm.Resource(
        name='res1',
        version='2.0.0',
        type=ocm.ArtefactType.BLOB,
        access=ocm.Access(),
    )

    examinee = create_upgrade_pr(
        from_ref=old_resource,
        to_ref=new_resource,
    )

    # test validation
    with pytest.raises(TypeError):
        examinee.target_matches(object()) # object is not of type DependencyBase

    # different type, same name and version
    assert not examinee.target_matches(
        ocm.Resource(
            name='res1',
            version='2.0.0',
            type=ocm.ArtefactType.OCI_IMAGE,
            access=None,
        )
    )

    # same type, and version, different name
    assert not examinee.target_matches(
        ocm.Resource(
            name='different-name',
            version='2.0.0',
            type=ocm.ArtefactType.BLOB,
            access=ocm.Access(),
        )
    )

    # same type, and name, different version
    assert not examinee.target_matches(
        ocm.Resource(
            name='res1',
            version='8.7.9',
            type=ocm.ArtefactType.BLOB,
            access=ocm.Access(),
        )
    )

    # all matches
    assert examinee.target_matches(
        ocm.Resource(
            name='res1',
            version='2.0.0',
            type=ocm.ArtefactType.BLOB,
            access=ocm.Access(),
        )
    )



import dataclasses
import typing


optional_str = typing.Optional[str]


@dataclasses.dataclass
class OperatingSystemId:
    '''
    Operating System identification, as specified in:
    https://www.freedesktop.org/software/systemd/man/os-release.html
    '''

    NAME: optional_str = None
    ID: optional_str = None
    PRETTY_NAME: optional_str = None
    CPE_NAME: optional_str = None
    VARIANT: optional_str = None
    VARIANT_ID: optional_str = None
    VERSION: optional_str = None
    VERSION_ID: optional_str = None
    VERSION_CODENAME: optional_str = None
    BUILD_ID: optional_str = None
    IMAGE_ID: optional_str = None
    IMAGE_VERSION: optional_str = None

    @property
    def is_distroless(self) -> bool:
        return self.PRETTY_NAME == 'Distroless'



import setuptools
import os

own_dir = os.path.abspath(os.path.dirname(__file__))


def requirements():
    yield 'gardener-oci>=' + version()

    with open(os.path.join(own_dir, 'requirements.txt')) as f:
        for line in f.readlines():
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            yield line


def modules():
    module_names = [
        os.path.basename(os.path.splitext(module)[0]) for module in
        os.scandir(path=own_dir)
        if module.is_file() and module.name.endswith('.py')
    ]

    # avoid including other setup-scripts
    module_names.remove('setup.oci')
    module_names.remove('setup.ocm')
    module_names.remove('setup.whd')
    return module_names


def packages():
    package_names = setuptools.find_packages()

    # remove packages (distributed via separate distribution-packages)
    package_names.remove('whd')
    package_names.remove('oci')
    #package_names.remove('ocm')
    return package_names


def version():
    with open(os.path.join(own_dir, 'ci', 'VERSION')) as f:
        return f.read().strip()


setuptools.setup(
    name='gardener-cicd-libs',
    version=version(),
    description='Gardener CI/CD Libraries',
    long_description='Gardener CI/CD Libraries',
    long_description_content_type='text/markdown',
    python_requires='>=3.10',
    py_modules=modules(),
    packages=packages(),
    package_data={
        '':['*.mako'],
        'ci': ['VERSION'],
        'concourse':[
            'resources/LAST_RELEASED_TAG',
            'resources/*.mako',
            '*.mako',
        ],
        'gci':[
            'ocm-component-descriptor-schema.yaml',
        ],
        'ocm':[
            'ocm-component-descriptor-schema.yaml',
        ],
    },
    install_requires=list(requirements()),
    entry_points={
    },
)



'''
utils used for replicating cnudie components between different repository contexts
'''


import dataclasses
import hashlib
import json
import logging

import gci.oci

import ccc.oci
import ocm
import oci
import oci.model as om

logger = logging.getLogger(__name__)


def replicate_oci_artifact_with_patched_component_descriptor(
    src_name: str,
    src_version: str,
    patched_component_descriptor: ocm.ComponentDescriptor,
    src_ctx_repo: ocm.OciOcmRepository,
):
    if isinstance(src_ctx_repo, str):
        src_ctx_repo = ocm.OciOcmRepository(baseUrl=src_ctx_repo)

    if not isinstance(src_ctx_repo, ocm.OciOcmRepository):
        raise NotImplementedError(src_ctx_repo)

    client = ccc.oci.oci_client()

    component = patched_component_descriptor.component
    target_repository = component.current_ocm_repo
    target_ref = target_repository.component_version_oci_ref(component)

    if client.head_manifest(image_reference=target_ref, absent_ok=True):
        # do not overwrite existing component-descriptors
        return

    src_ref = src_ctx_repo.component_version_oci_ref(
        name=src_name,
        version=src_version,
    )

    src_manifest = client.manifest(
        image_reference=src_ref,
    )

    raw_fobj = gci.oci.component_descriptor_to_tarfileobj(patched_component_descriptor)

    cd_digest = hashlib.sha256()
    while (chunk := raw_fobj.read(4096)):
        cd_digest.update(chunk)

    cd_octets = raw_fobj.tell()
    cd_digest = cd_digest.hexdigest()
    cd_digest_with_alg = f'sha256:{cd_digest}'
    raw_fobj.seek(0)

    # src component descriptor OciBlobRef for patching
    src_config_dict = json.loads(client.blob(src_ref, src_manifest.config.digest).content)
    src_component_descriptor_oci_blob_ref = om.OciBlobRef(
        **src_config_dict['componentDescriptorLayer'],
    )

    # config OciBlobRef
    cfg = gci.oci.ComponentDescriptorOciCfg(
        componentDescriptorLayer=gci.oci.ComponentDescriptorOciBlobRef(
            digest=cd_digest_with_alg,
            size=cd_octets,
        ),
    )
    cfg_raw = json.dumps(dataclasses.asdict(cfg)).encode('utf-8')

    # replicate all blobs except overwrites
    target_manifest = oci.replicate_blobs(
        src_ref=src_ref,
        src_oci_manifest=src_manifest,
        tgt_ref=target_ref,
        oci_client=client,
        blob_overwrites={
            src_component_descriptor_oci_blob_ref: raw_fobj,
            src_manifest.config: cfg_raw,
        },
    )

    target_manifest_dict = dataclasses.asdict(target_manifest)
    target_manifest_bytes = json.dumps(target_manifest_dict).encode('utf-8')

    client.put_manifest(
        image_reference=target_ref,
        manifest=target_manifest_bytes,
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import os

import mako.template
import mako.lookup

import makoutil

steps_dir = os.path.abspath(os.path.dirname(__file__))
template_lookup = mako.lookup.TemplateLookup(directories=(steps_dir,))


def step_template(name):
    with makoutil.template_lock:
        return template_lookup.get_template(f'/{name}.mako')


def step_def(name):
    template = step_template(name)

    return template.get_def(name + '_step').render


def step_lib_def(name):
    template = step_template(name)

    return template.get_def(name + '_step_lib').render


def step_lib(name):
    module_file = os.path.join(steps_dir, name + '.py')
    with open(module_file) as f:
        return f.read()



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import enum

from concourse.model.job import (
    JobVariant,
)
from concourse.model.step import (
    PipelineStep,
    PullRequestNotificationPolicy,
    StepNotificationPolicy,
)
from concourse.model.base import (
  AttributeSpec,
  Trait,
  TraitTransformer,
  ScriptType,
  ModelValidationError
)


class VersionInterface(enum.Enum):
    FILE = 'file'
    CALLBACK = 'callback'


ATTRIBUTES = (
    AttributeSpec.optional(
        name='preprocess',
        default='inject-commit-hash',
        doc='sets the semver version operation to calculate the effective version during the build',
    ),
    AttributeSpec.optional(
        name='versionfile',
        default='VERSION',
        doc='relative path to the version file',
    ),
    AttributeSpec.optional(
        name='inject_effective_version',
        default=False,
        doc='''
        whether or not the effective version is to be written into the source tree's VERSION file
        ''',
        type=bool,
    ),
    AttributeSpec.optional(
        name='version_interface',
        default=VersionInterface.FILE,
        doc='''\
        how the version can be read/written. This is done automatically set to
        "callback", if "read_callback" and "write_callback" are set. Only here
        for compatibility reasons.
        ''',
        type=VersionInterface,
    ),
    AttributeSpec.optional(
        name='read_callback',
        default=None,
        doc='relative path to an executable that returns current version via stdout',
    ),
    AttributeSpec.optional(
        name='write_callback',
        default=None,
        doc='relative path to an executable that accepts version from stdin and writes it',
    ),
)


class VersionTrait(Trait):
    PREPROCESS_OPS = {
        'finalise',
        'finalize',
        'finalise-skip-patchlevel-zero',
        'finalize-skip-patchlevel-zero',
        'inject-branch-name',
        'inject-commit-hash',
        'inject-commit-hash-nodash',
        'inject-timestamp',
        'noop',
        'use-branch-name',
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    @property
    def preprocess(self):
        return self.raw['preprocess']

    def versionfile_relpath(self):
        return self.raw['versionfile']

    def inject_effective_version(self):
        return self.raw['inject_effective_version']

    def version_interface(self):
        return VersionInterface(self.raw.get('version_interface'))

    def read_callback(self):
        return self.raw.get('read_callback')

    def write_callback(self):
        return self.raw.get('write_callback')

    def transformer(self):
        return VersionTraitTransformer(trait=self)

    def validate(self):
        super().validate()

        if not self.preprocess in self.PREPROCESS_OPS:
            raise ValueError('preprocess must be one of: ' + ', '.join(self.PREPROCESS_OPS))

        if self.read_callback() and (not self.write_callback()) or \
          (not self.read_callback()) and self.write_callback():
            raise ModelValidationError(
                f"{self.write_callback()=}' {self.read_callback()=}'. Set either both or none!"
            )


ENV_VAR_NAME = 'version_path'
DIR_NAME = 'managed-version'


class VersionTraitTransformer(TraitTransformer):
    name = 'version'

    def __init__(self, trait: VersionTrait):
        super().__init__()

        # Set version_interface to 'callback' if write_callback and read_callback are set
        if trait.read_callback() and trait.write_callback():
            trait.raw['version_interface'] = 'callback'

    def inject_steps(self):
        self.version_step = PipelineStep(
            name='version',
            raw_dict={},
            is_synthetic=True,
            notification_policy=StepNotificationPolicy.NO_NOTIFICATION,
            pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
            injecting_trait_name=self.name,
            script_type=ScriptType.PYTHON3,
            )
        self.version_step.add_output(name=DIR_NAME, variable_name=ENV_VAR_NAME)
        self.version_step.set_timeout(duration_string='5m')

        yield self.version_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        # all steps depend from us and may consume our output
        for step in pipeline_args.steps():
            if step == self.version_step:
                continue
            step._add_dependency(self.version_step)
            step.add_input(variable_name=ENV_VAR_NAME, name=DIR_NAME)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

from model.base import (
    NamedModelElement,
)


class IngressConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''
    def tls_host_names(self):
        return self.raw['tls_host_names']

    def ttl(self):
        return self.raw['ttl']

    def issuer_name(self):
        return self.raw['issuer_name']

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'issuer_name',
            'tls_host_names',
            'ttl',
        ]



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import enum
import re
import typing

import dacite

from concourse.model.job import (
    JobVariant,
)
from concourse.model.step import (
    PipelineStep,
    PullRequestNotificationPolicy,
)
from concourse.model.base import (
    AttributeSpec,
    Trait,
    TraitTransformer,
    ModelBase,
    ScriptType,
)
import github.compliance.model as gcm
from model.base import ModelValidationError

import concourse.model.traits.component_descriptor


OS_ID_SCAN_ATTRS = (
    AttributeSpec.optional(
        name='parallel_jobs',
        default=8,
        doc='amount of parallel jobs to run',
        type=int,
    ),
    AttributeSpec.optional(
        name='timeout',
        default='2h',
        doc='''
        go-style time interval (e.g.: '1h30m') after which the image-scan-step will be interrupted
        and fail.
        ''',
        type=str,
    ),
)


class OsIdScan(ModelBase):
    @classmethod
    def _attribute_specs(cls):
        return OS_ID_SCAN_ATTRS

    def parallel_jobs(self) -> int:
        return int(self.raw['parallel_jobs'])

    def timeout(self):
        return self.raw.get('timeout')


class Notify(enum.Enum):
    EMAIL_RECIPIENTS = 'email_recipients'
    NOBODY = 'nobody'
    COMPONENT_OWNERS = 'component_owners'
    GITHUB_ISSUES = 'github_issues'


@dataclasses.dataclass
class GithubIssueTemplateCfg:
    body: str
    type: str


@dataclasses.dataclass
class LicenseCfg:
    '''
    configures license policies for discovered licences

    licenses are configured as lists of regular expressions (matching is done case-insensitive)
    '''
    prohibited_licenses: typing.Optional[list[str]] = None

    def is_allowed(self, license: str):
        if not self.prohibited_licenses:
            return True

        for prohibited in self.prohibited_licenses:
            if re.fullmatch(prohibited, license, re.IGNORECASE):
                return False
        else:
            return True


@dataclasses.dataclass(frozen=True)
class IssuePolicies:
    max_processing_time_days: gcm.MaxProcessingTimesDays = gcm.MaxProcessingTimesDays()


ATTRIBUTES = (
    AttributeSpec.optional(
        name='notify',
        default=Notify.EMAIL_RECIPIENTS,
        doc='whom to notify about found issues',
        type=Notify,
    ),
    AttributeSpec.optional(
        name='issue_policies',
        default=IssuePolicies(),
        type=IssuePolicies,
        doc='defines issues policies (e.g. SLAs for maximum processing times',
    ),
    AttributeSpec.optional(
        name='overwrite_github_issues_tgt_repository_url',
        default=None,
        doc='if set, and notify is set to github_issues, overwrite target github repository',
    ),
    AttributeSpec.optional(
        name='github_issue_templates',
        default=None,
        doc='''\
        use to configure custom github-issue-templates (sub-attr: `body`)
        use python3's format-str syntax

        .. code-block::
          :caption: available variables

          - summary # contains name, version, etc in a table
          - component_name
          - component_version
          - resource_name
          - resource_version
          - resource_type
          - greatest_cve
          - report_url
          - delivery_dashboard_url

        ''',
        type=list[GithubIssueTemplateCfg],
    ),
    AttributeSpec.optional(
        name='github_issue_labels_to_preserve',
        default=None,
        doc='optional list of regexes for labels that will never be removed upon ticket-update',
        type=list[str],
    ),
    AttributeSpec.optional(
        name='email_recipients',
        default=[],
        doc='optional list of email recipients to be notified about critical scan results',
        type=typing.List[str],
    ),
    AttributeSpec.optional(
        name='os_id',
        default=None,
        type=OsIdScan,
        doc='if present, identify operating system',
    ),
    AttributeSpec.optional(
        name='trait_depends',
        default=(),
        type=typing.Set[str],
        doc='if present, generated build steps depend on those generated from specified traits',
    ),
)


class ImageScanTrait(Trait):
    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def issue_policies(self) -> IssuePolicies:
        if isinstance((v := self.raw['issue_policies']), IssuePolicies):
            return v

        return dacite.from_dict(
            data_class=IssuePolicies,
            data=v,
        )

    def notify(self) -> Notify:
        return Notify(self.raw['notify'])

    def overwrite_github_issues_tgt_repository_url(self) -> typing.Optional[str]:
        return self.raw.get('overwrite_github_issues_tgt_repository_url')

    def github_issue_templates(self) -> list[GithubIssueTemplateCfg]:
        if not (raw := self.raw.get('github_issue_templates')):
            return None

        template_cfgs = [
            dacite.from_dict(
                data_class=GithubIssueTemplateCfg,
                data=cfg,
            ) for cfg in raw
        ]

        return template_cfgs

    def github_issue_template(self, type: str) -> typing.Optional[GithubIssueTemplateCfg]:
        if not (template_cfgs := self.github_issue_templates()):
            return None

        for cfg in template_cfgs:
            if cfg.type == type:
                return cfg

        return None

    def github_issue_labels_to_preserve(self) -> typing.Optional[list[str]]:
        return self.raw['github_issue_labels_to_preserve']

    def email_recipients(self):
        return self.raw['email_recipients']

    def os_id(self):
        if (raw := self.raw.get('os_id')) is not None:
            return OsIdScan(raw_dict=raw)

    def transformer(self):
        return ImageScanTraitTransformer(trait=self)

    def trait_depends(self):
        return set(self.raw['trait_depends'])

    def validate(self):
        super().validate()


class ImageScanTraitTransformer(TraitTransformer):
    name = 'image_scan'

    def __init__(self, trait, *args, **kwargs):
        self.trait = trait
        super().__init__(*args, **kwargs)

    def inject_steps(self):
        if self.trait.os_id():
            self.os_id_step = PipelineStep(
                    name='os-id-scan',
                    raw_dict={},
                    is_synthetic=True,
                    pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
                    injecting_trait_name=self.name,
                    script_type=ScriptType.PYTHON3
            )
            self.os_id_step.add_input(
                name=concourse.model.traits.component_descriptor.DIR_NAME,
                variable_name=concourse.model.traits.component_descriptor.ENV_VAR_NAME,
            )
            self.os_id_step.set_timeout(duration_string=self.trait.os_id().timeout())
            yield self.os_id_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        # our steps depends on dependency descriptor step
        component_descriptor_step = pipeline_args.step(
            concourse.model.traits.component_descriptor.DEFAULT_COMPONENT_DESCRIPTOR_STEP_NAME
        )
        if self.trait.os_id():
            self.os_id_step._add_dependency(component_descriptor_step)

        for trait_name in self.trait.trait_depends():
            if not pipeline_args.has_trait(trait_name):
                raise ModelValidationError(f'dependency towards absent trait: {trait_name}')

    @classmethod
    def dependencies(cls):
        return {'component_descriptor'}

    @classmethod
    def order_dependencies(cls):
        # required in case image-scanning should be done after publish
        # (-> auto-injected dependency for "prepare"-step towards _all_ steps)
        return {'publish'}



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import concourse.util as examinee


def test_static_pipeline_metadata(monkeypatch):
    # fake running on ci
    monkeypatch.setenv('CC_ROOT_DIR', 'made/up/dir')

    TEST_CONFIG_SET_NAME = 'made-up_config_name'
    TEST_CONCOURSE_TEAM_NAME = 'made-up_concourse_team'
    TEST_PIPELINE_NAME = 'made-up_pipeline_name'
    TEST_JOB_NAME = 'made-up_job_name'

    test_metadata = examinee.PipelineMetaData(
        pipeline_name=TEST_PIPELINE_NAME,
        job_name=TEST_JOB_NAME,
        current_config_set_name=TEST_CONFIG_SET_NAME,
        team_name=TEST_CONCOURSE_TEAM_NAME,
    )

    monkeypatch.setenv('CONCOURSE_CURRENT_CFG', TEST_CONFIG_SET_NAME)
    monkeypatch.setenv('CONCOURSE_CURRENT_TEAM', TEST_CONCOURSE_TEAM_NAME)
    monkeypatch.setenv('PIPELINE_NAME', TEST_PIPELINE_NAME)
    monkeypatch.setenv('BUILD_JOB_NAME', TEST_JOB_NAME)

    pipeline_metadata = examinee.get_pipeline_metadata()

    assert pipeline_metadata == test_metadata



# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import dataclasses
import typing

import ci.util
import ctt.processing_model as pm
import ctt.util as ctt_util
import ocm
import oci.client
import oci.model as om

original_ref_label_name = 'cloud.gardener.cnudie/migration/original_ref'


class IdentityUploader:
    def process(self, processing_job, target_as_source=False):
        upload_request = processing_job.upload_request

        _, _, src_tag = oci.client._split_image_reference(upload_request.source_ref)
        if ':' in src_tag:
            raise NotImplementedError

        if not target_as_source:
            upload_request = dataclasses.replace(
                processing_job.upload_request,
                target_ref=processing_job.upload_request.source_ref,
            )

        return dataclasses.replace(
            processing_job,
            upload_request=upload_request,
        )


def labels_with_migration_hint(
    resource: ocm.Resource,
    src_img_ref,
):
    original_ref_label = ocm.Label(
        name=original_ref_label_name,
        value=src_img_ref,
    )
    src_labels = resource.labels or []
    return ctt_util.add_label(
        src_labels=src_labels,
        label=original_ref_label,
    )


class PrefixUploader:
    def __init__(
        self,
        prefix,
        mangle=True,
        mangle_replacement_char='_',
        convert_to_relative_refs=False,
        remove_prefixes=[],
        **kwargs
    ):
        super().__init__(**kwargs)

        self._prefix = prefix
        self._mangle = mangle
        self._mangle_replacement_char = mangle_replacement_char
        self._convert_to_relative_refs = convert_to_relative_refs
        self._remove_prefixes = remove_prefixes

    def process(
        self,
        processing_job: pm.ProcessingJob,
        target_as_source=False,
    ):
        if processing_job.resource.access.type is not ocm.AccessType.OCI_REGISTRY:
            raise RuntimeError('PrefixUploader only supports access type == ociRegistry')

        if not target_as_source:
            src_ref = processing_job.resource.access.imageReference
        else:
            src_ref = processing_job.upload_request.target_ref

        src_ref = om.OciImageReference.to_image_ref(src_ref)
        src_base_ref = src_ref.ref_without_tag

        for remove_prefix in self._remove_prefixes:
            src_base_ref = src_base_ref.removeprefix(remove_prefix)

        if self._mangle:
            src_base_ref = src_base_ref.replace('.', self._mangle_replacement_char)

        # if a prefix is to be removed from existing src base ref, it is likely that it should be
        # replaced by the new prefix, instead of only prepended (where a joining `/` is reasonable).
        # Instead, leave it up to the configuration to decide on the joining character.
        if not self._remove_prefixes:
            tgt_ref = ci.util.urljoin(
                self._prefix,
                src_base_ref,
            )
        else:
            tgt_ref = self._prefix + src_base_ref

        if src_ref.has_mixed_tag:
            symbolical_tag, digest_tag = src_ref.parsed_mixed_tag
            tgt_ref = f'{tgt_ref}:{symbolical_tag}@{digest_tag}'
        elif src_ref.has_digest_tag:
            tgt_ref = f'{tgt_ref}@{src_ref.tag}'
        elif src_ref.has_symbolical_tag:
            tgt_ref = f'{tgt_ref}:{src_ref.tag}'

        if src_ref.has_digest_tag:
            processing_job.upload_request.reference_target_by_digest = True

        upload_request = dataclasses.replace(
            processing_job.upload_request,
            source_ref=processing_job.resource.access.imageReference,
            target_ref=tgt_ref,
        )

        if self._convert_to_relative_refs:
            # remove host from target ref
            # don't use artifact_path as self._prefix can also contain path elements
            relative_artifact_path = om.OciImageReference.to_image_ref(
                image_reference=tgt_ref,
                normalise=False, # don't inject docker special handlings
            ).local_ref
            access = ocm.RelativeOciAccess(
                reference=relative_artifact_path,
            )
        else:
            access = ocm.OciAccess(
                imageReference=tgt_ref,
            )

        # propagate changed resource
        processing_job.processed_resource = dataclasses.replace(
            processing_job.resource,
            access=access,
            labels=labels_with_migration_hint(
                resource=processing_job.resource,
                src_img_ref=processing_job.resource.access.imageReference,
            ),
        )

        return dataclasses.replace(
            processing_job,
            upload_request=upload_request
        )


class TagSuffixUploader:
    def __init__(
        self,
        suffix,
        separator='-',
    ):
        self._suffix = suffix
        self._separator = separator

    def process(
        self,
        processing_job: pm.ProcessingJob,
        target_as_source: bool=False,
    ):
        if processing_job.resource.access.type is not ocm.AccessType.OCI_REGISTRY:
            raise NotImplementedError

        if not target_as_source:
            src_ref = processing_job.resource.access.imageReference
        else:
            src_ref = processing_job.upload_request.target_ref

        src_ref = om.OciImageReference.to_image_ref(src_ref)

        if src_ref.has_digest_tag:
            raise RuntimeError('Cannot append tag suffix to resource that is accessed via digest')

        tgt_tag = self._separator.join((src_ref.tag, self._suffix))
        tgt_ref = ':'.join((src_ref.ref_without_tag, tgt_tag))

        upload_request = dataclasses.replace(
            processing_job.upload_request,
            source_ref=processing_job.resource.access.imageReference,
            target_ref=tgt_ref,
        )

        # propagate changed resource
        processing_job.processed_resource = dataclasses.replace(
            processing_job.resource,
            access=ocm.OciAccess(
                imageReference=tgt_ref,
            ),
            labels=labels_with_migration_hint(
                resource=processing_job.resource,
                src_img_ref=processing_job.resource.access.imageReference,
            ),
        )

        return dataclasses.replace(
            processing_job,
            upload_request=upload_request
        )


class ExtraTagUploader:
    '''
    Uploader that will push additional (static) tags to uploaded images. Useful to e.g. add
    `latest` tag. Extra-Tags will be overwritten as a hardcoded behaviour of this uploader.
    '''
    def __init__(self, extra_tags: typing.Iterable[str]):
        self.extra_tags = tuple(extra_tags)

    def process(self, processing_job, target_as_source=False):
        processing_job.extra_tags = self.extra_tags

        return processing_job


class DigestUploader:
    '''
    sets `reference_target_by_digest` attribute in upload-request, which will result in
    target-component-descriptor's resouce's access use digest rather than tag to reference
    oci image. If `retain_symbolic_tag` is set, the symbolic tag is kept and the digest
    is appended, otherwise the digest overwrites the symbolic tag.
    '''
    def __init__(
        self,
        retain_symbolic_tag: bool=False,
    ):
        self._retain_symbolic_tag = retain_symbolic_tag

    def process(self, processing_job: pm.ProcessingJob, target_as_source=False):
        processing_job.upload_request.reference_target_by_digest = True
        processing_job.upload_request.retain_symbolic_tag = self._retain_symbolic_tag

        return processing_job



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import dataclasses
import json
import logging
import time
import typing

import dateutil.parser
import requests

from dacite import from_dict
from dacite.exceptions import DaciteError
from contextlib import contextmanager
from dataclasses import dataclass
from datetime import (
    datetime,
    timedelta,
)

import concourse.client.api
from whd.model import (
    PullRequestEvent,
)
from .model import (
    BuildStatus,
    Job,
    PipelineConfigResource,
    PipelineResource,
    ResourceType,
    ResourceVersion,
    PinnedPRVersion,
    PinnedGitVersion,
    PinnedTimeVersion,
)


logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


class PinningUnnecessary(Exception):
    pass


class PinningFailedError(Exception):
    pass


@dataclass
class PinComment:
    pin_timestamp: str
    version: typing.Union[PinnedPRVersion, PinnedGitVersion, PinnedTimeVersion]
    next_retry: str
    comment: str


def determine_pr_resource_versions(
    pr_event: PullRequestEvent,
    concourse_api,
) -> typing.Tuple[PipelineConfigResource, typing.Sequence[ResourceVersion]]:

    pr_resources = concourse_api.pipeline_resources(
        pipeline_names=concourse_api.pipelines(),
        resource_type=ResourceType.PULL_REQUEST,
    )

    for pr_resource in pr_resources:
        ghs = pr_resource.github_source()
        pr_repository = pr_event.repository()
        if not ghs.hostname() == pr_repository.github_host():
            continue
        if not ghs.repo_path().lstrip('/') == pr_repository.repository_path():
            continue

        pipeline_name = pr_resource.pipeline.name
        pr_resource_versions = concourse_api.resource_versions(
            pipeline_name=pipeline_name,
            resource_name=pr_resource.name,
        )

        # only interested in pr resource versions, which match the PR number
        pr_resource_versions = [
            rv for rv in pr_resource_versions if rv.version()['pr'] == str(pr_event.number())
        ]

        yield (pr_resource, pr_resource_versions)


def determine_jobs_to_be_triggered(
    *resources: PipelineConfigResource,
) -> typing.Iterator[Job]:

    for resource in resources:
        yield from (
            job for job in resource.pipeline.jobs() if job.is_triggered_by_resource(resource.name)
        )


def wait_for_job_to_be_triggered(
    job: Job,
    resource_version: ResourceVersion,
    concourse_api,
    retries: int=10,
    sleep_time_seconds: int=6,
) -> bool:
    '''
    There is some delay between the update of a resource and the start of the corresponding job.
    Therefore we have to wait some time to decide if the job has been triggered or not.

    @return True if the job has been triggered
    '''
    while retries >= 0:
        if has_job_been_triggered(
            job=job,
            resource_version=resource_version,
            concourse_api=concourse_api,
        ):
            return True

        time.sleep(sleep_time_seconds)
        retries -= 1

    return False


def has_job_been_triggered(
    job: Job,
    resource_version: ResourceVersion,
    concourse_api,
) -> bool:

    builds = concourse_api.job_builds(job.pipeline.name, job.name)
    for build in builds[-20:]:
        if build.status() == BuildStatus.PENDING:
            # cannot retrieve build plan for a pending job
            continue
        if build.status() in [
            BuildStatus.ABORTED,
            BuildStatus.ERRORED,
            BuildStatus.FAILED,
        ]:
            # don't consider failed jobs
            continue

        try:
            build_plan = build.plan()
        except requests.exceptions.HTTPError as e:
            # a job that just now went into running state could still have no build plan
            if e.response.status_code is not requests.status_codes.codes.NOT_FOUND: # noqa
                raise e
            continue
        if build_plan.contains_version_ref(resource_version.version()['ref']):
            return True
    return False


def jobs_not_triggered(
    pr_event: PullRequestEvent,
    concourse_api,
) -> typing.Tuple[Job, PipelineConfigResource, ResourceVersion]:

    if not (head_commit := pr_event.head_commit()):
        logger.warn(
            f'Unable to determine head-commit-sha of PR #{pr_event.number()} for '
            f'repository {pr_event.repository().repository_path()} (from {pr_event.delivery()=}). '
            'Jobs will be triggered for every commit of the PR.'
        )

    for pr_resource, pr_resource_versions in determine_pr_resource_versions(
        pr_event=pr_event,
        concourse_api=concourse_api,
    ):
        for job in determine_jobs_to_be_triggered(pr_resource):
            for pr_resource_version in pr_resource_versions:

                # we only run jobs for the head-commit of the PR (if it can be determined)
                if (
                    head_commit
                    and pr_resource_version.version()['ref'] != head_commit
                ):
                    continue

                if not wait_for_job_to_be_triggered(job, pr_resource_version, concourse_api):
                    yield (job, pr_resource, pr_resource_version)
                else:
                    logger.info(
                        f'Resource version {pr_resource_version.version()} '
                        f'triggered {job.name=} of {pr_resource.pipeline_name()=} '
                        '- no pinning necessary'
                    )


def pin_resource_and_trigger_build(
    job: Job,
    resource: PipelineConfigResource,
    resource_version: ResourceVersion,
    concourse_api,
    retries: int=3,
):
    retries_for_job_trigger = 10
    sleep_seconds_between_attempts = 6
    wait_until_next_retry_seconds = retries_for_job_trigger * sleep_seconds_between_attempts

    with ensure_resource_pinned(
        resource=resource,
        resource_version=resource_version,
        wait_until_next_retry_seconds=wait_until_next_retry_seconds,
        concourse_api=concourse_api,
    ):

        for count in range(retries):
            concourse_api.trigger_build(
                pipeline_name=resource.pipeline_name(),
                job_name=job.name,
            )
            if wait_for_job_to_be_triggered(
                job=job,
                resource_version=resource_version,
                concourse_api=concourse_api,
                retries=retries_for_job_trigger,
                sleep_time_seconds=sleep_seconds_between_attempts,
            ):
                logger.info(f'{job.name=} for {resource_version.version()=} has been triggered')
                return

            else:
                # job did not start, update pin comment with new retry timestamp
                _pin_and_comment_resource(
                    wait_until_next_retry_seconds=wait_until_next_retry_seconds,
                    resource=resource,
                    resource_version=resource_version,
                    concourse_api=concourse_api,
                )
                logger.warning(
                    f'job {job.name} for resource version {resource_version.version()} '
                    f'could not be triggered. Retry {count + 1}/{retries}'
                )

        logger.warning(
            f'job {job.name} for resource version {resource_version.version()} '
            f'could not be triggered. Giving up after {retries} retries'
        )


def _pin_and_comment_resource(
    wait_until_next_retry_seconds: int,
    resource: PipelineConfigResource,
    resource_version: ResourceVersion,
    concourse_api: concourse.client.api.ConcourseApiBase,
):
    now = datetime.now()
    next_retry = now + timedelta(seconds=wait_until_next_retry_seconds)
    pin_comment = PinComment(
        pin_timestamp=now.isoformat(),
        version=resource_version.version(),
        next_retry=next_retry.isoformat(),
        comment=(
            "pinned by technical user concourse, please do not unpin before "
            f"{next_retry.isoformat()}"
        ),
    )
    concourse_api.pin_and_comment_resource_version(
        pipeline_name=resource.pipeline_name(),
        resource_name=resource.name,
        resource_version_id=resource_version.id(),
        comment=json.dumps(dataclasses.asdict(pin_comment)),
    )
    logger.info(f'pinned resource: {resource.name} version: {resource_version.version()}')


@contextmanager
def ensure_resource_pinned(
    resource: PipelineResource,
    resource_version: ResourceVersion,
    wait_until_next_retry_seconds: int,
    concourse_api,
):
    try:
        _ensure_resource_unpinned(
            resource=resource,
            resource_version=resource_version,
            concourse_api=concourse_api,
        )
        # pin resource and add comment
        _pin_and_comment_resource(
            wait_until_next_retry_seconds=wait_until_next_retry_seconds,
            resource=resource,
            resource_version=resource_version,
            concourse_api=concourse_api,
        )
        yield
    finally:
        if concourse_api.unpin_resource(
            pipeline_name=resource.pipeline_name(),
            resource_name=resource.name,
        ):
            logger.info(f'successfully unpinned {resource.name=}')
        else:
            logger.info(f'{resource.name=} not pinned - no unpinning necessary')


def _ensure_resource_unpinned(
    resource: PipelineResource,
    resource_version: ResourceVersion,
    concourse_api,
    maximal_checktime_seconds: int = 240,
):
    # maximal time to wait for pinned resource, after that we give up
    latest_unpin_checktime = datetime.now() + timedelta(seconds=maximal_checktime_seconds)

    while latest_unpin_checktime >= datetime.now():
        cc_resource = concourse_api.resource(
            pipeline_name=resource.pipeline_name(),
            resource_name=resource.name,
        )

        if not cc_resource.is_pinned():
            return

        sleep_time = 60
        try:
            pin_comment = from_dict(
                data_class=PinComment,
                data=json.loads(cc_resource.pin_comment()),
            )

            if pin_comment.version.ref == resource_version.version()['ref']:
                raise PinningUnnecessary(
                    "Resource is already pinned for the same version. Not waiting for unpin"
                )

            next_retry_time = dateutil.parser.isoparse(pin_comment.next_retry)
            sleep_time = min(
                (next_retry_time - datetime.now()).seconds + 5,
                maximal_checktime_seconds,
            )
            if next_retry_time <= datetime.now():
                logger.warning(
                    'found resource which was not unpinned although timestamp expired: '
                    f'{resource.pipeline_name()=} {resource.name=}. Unpinning resource now'
                )
                concourse_api.unpin_resource(
                    pipeline_name=resource.pipeline_name(),
                    resource_name=resource.name,
                )
                return

        except json.JSONDecodeError:
            logger.warning(
                f'Resource comment for {resource.name=} of {resource.pipeline_name()=} '
                'could not be parsed - most likely pinned by human user'
            )
        except DaciteError:
            logger.warning(
                f'Resource comment for {resource.name=} of {resource.pipeline_name()=} '
                'could not be instantiated'
            )

        logger.info(
            f'{resource.name=} of {resource.pipeline_name()=} is logged by another thread. '
            f'Sleep for {sleep_time}s and try again'
        )
        time.sleep(sleep_time)

    raise PinningFailedError(
        f'Tried at least {maximal_checktime_seconds} seconds to unpin '
        f'{resource.name=} of pipeline {resource.pipeline_name()=}'
    )



'''
This module comprises model classes to define JSON Web Keys according
to https://datatracker.ietf.org/doc/html/rfc7518 as well as convenience
functions to it.
'''
import base64
import dataclasses
import datetime
import enum
import functools
import logging
import typing

import Crypto.PublicKey.RSA
import Crypto.Util.number
import dacite
import jwt

import model.delivery

logger = logging.getLogger(__name__)
JWT_KEY = 'bearer_token'


class KeyType(enum.StrEnum):
    RSA = 'RSA'
    OCTET_SEQUENCE = 'oct' # used to represent symmetric keys


class Use(enum.StrEnum):
    SIGNATURE = 'sig'
    ENCRYPTION = 'enc'


class Algorithm(enum.StrEnum):
    RS256 = 'RS256'
    HS256 = 'HS256'


@dataclasses.dataclass(frozen=True)
class JSONWebKey:
    kty: KeyType
    use: Use
    alg: Algorithm
    kid: str

    @property
    def key(self) -> bytes:
        # has to be defined by derived classes
        raise NotImplementedError

    @staticmethod
    def from_dict(data: dict) -> typing.Self:
        algorithm = Algorithm(data.get('alg'))

        class_by_algorithm = {
            Algorithm.RS256: RSAPublicKey,
            Algorithm.HS256: SymmetricKey,
        }

        return dacite.from_dict(
            data_class=class_by_algorithm.get(algorithm),
            data=data,
            config=dacite.Config(
                cast=[enum.Enum],
            ),
        )

    @staticmethod
    def from_signing_cfg(signing_cfg: model.delivery.SigningCfg) -> typing.Self:
        algorithm = Algorithm(signing_cfg.algorithm().upper())
        use = Use.SIGNATURE
        kid = signing_cfg.id()

        if algorithm == Algorithm.RS256:
            public_key = Crypto.PublicKey.RSA.import_key(signing_cfg.public_key())

            return RSAPublicKey(
                use=use,
                kid=kid,
                n=encodeBase64urlUInt(public_key.n),
                e=encodeBase64urlUInt(public_key.e),
            )
        elif algorithm == Algorithm.HS256:
            return SymmetricKey(
                use=use,
                kid=kid,
                k=encodeBase64url(signing_cfg.secret().encode('utf-8')),
            )


@dataclasses.dataclass(frozen=True, kw_only=True)
class RSAPublicKey(JSONWebKey):
    n: str # modulus (Base64urlUInt encoded)
    e: str # exponent (Base64urlUInt encoded)
    kty: KeyType = KeyType.RSA
    alg: Algorithm = Algorithm.RS256

    @property
    def key(self) -> bytes:
        return Crypto.PublicKey.RSA.construct(
            rsa_components=(
                decodeBase64urlUInt(self.n),
                decodeBase64urlUInt(self.e),
            ),
        ).export_key(format='PEM')


@dataclasses.dataclass(frozen=True, kw_only=True)
class SymmetricKey(JSONWebKey):
    k: str # key value (base64url encoded)
    kty: KeyType = KeyType.OCTET_SEQUENCE
    alg: Algorithm = Algorithm.HS256

    @property
    def key(self) -> bytes:
        return decodeBase64url(self.k)


def encodeBase64url(b: bytes) -> str:
    url_encoding = base64.urlsafe_b64encode(b)

    return url_encoding.rstrip(b'=').decode('utf-8')


def decodeBase64url(s: str) -> bytes:
    return base64.urlsafe_b64decode(s + '==')


def encodeBase64urlUInt(number: int) -> str:
    b = Crypto.Util.number.long_to_bytes(number)

    return encodeBase64url(b)


def decodeBase64urlUInt(s: str) -> int:
    b = decodeBase64url(s)

    return Crypto.Util.number.bytes_to_long(b)


@functools.cache
def decode_jwt(
    token: str,
    verify_signature: bool=True,
    json_web_key: JSONWebKey | None=None,
    **kwargs,
) -> dict:
    '''
    This is just a convenience wrapper for `jwt.decode` which eases
    signature validation of a given `token` using a `JSONWebKey`.
    '''
    if verify_signature and not json_web_key:
        raise ValueError('`json_web_key` must be specified if `verify_signature` is `True`')

    return jwt.decode(
        jwt=token,
        key=json_web_key.key if json_web_key else None,
        algorithms=[json_web_key.alg if json_web_key else None],
        options={
            'verify_signature': verify_signature,
        },
        **kwargs,
    )


def is_jwt_token_expired(
    token: str,
) -> bool:
    decoded_jwt = decode_jwt(
        token=token,
        verify_signature=False,
    )

    expiration_date = datetime.datetime.fromtimestamp(
        timestamp=decoded_jwt.get('exp'),
        tz=datetime.timezone.utc,
    )

    now = datetime.datetime.now(tz=datetime.timezone.utc)

    return now > expiration_date



import dataclasses
import functools
import logging
import typing
from collections import defaultdict

import ocm
import release_notes.model as rnm

logger = logging.getLogger(__name__)


@functools.total_ordering
@dataclasses.dataclass
class Title:
    display: str
    identifiers: list[str]
    priority: int

    def __hash__(self):
        return hash(self.display)

    def __eq__(self, other):
        return (
            other is not None
            and isinstance(other, Title)
            and (other == self or hash(other) == hash(self))
        )

    def __lt__(self, other):
        if not isinstance(other, Title):
            raise ValueError(other)
        return self.priority < other.priority


def _simple_title(display: str) -> Title:
    return Title(display=display, identifiers=[display.lower()], priority=0)


categories = [
    Title(display='⚠️ Breaking Changes', identifiers=['action', 'breaking'], priority=0),
    Title(display='📰 Noteworthy', identifiers=['noteworthy'], priority=1),
    Title(display='✨ New Features', identifiers=['feature'], priority=2),
    Title(display='🐛 Bug Fixes', identifiers=['bugfix', 'fix'], priority=3),
    Title(display='🏃 Others', identifiers=['improvement', 'other'], priority=4),
    Title(display='📖 Documentation', identifiers=['doc'], priority=5),
]

target_groups = [
    _simple_title('USER'),
    _simple_title('OPERATOR'),
    _simple_title('DEVELOPER'),
    _simple_title('DEPENDENCY'),
]

# identifier -> Title
categories_by_identifier: dict[str, Title] = {i: k for k in categories for i in k.identifiers}
target_groups_by_identifier: dict[str, Title] = {i: k for k in target_groups for i in k.identifiers}


@dataclasses.dataclass
class Header:
    level: int
    title: str

    def __str__(self):
        return f"{'#' * self.level} {self.title}\n"  # there should be a new line after the header


@dataclasses.dataclass
class ListItem:
    level: int
    text: str

    def __str__(self):
        return f"{'  ' * (self.level - 1)}- {self.text}"


def list_item_from_lines(lines: list[str], start_level: int = 1) -> list[ListItem]:
    objs = []

    level = start_level
    last_index: typing.Optional[int] = None
    for idx, line in enumerate(lines):
        if line.startswith(' ') or line.startswith('\t'):
            first_index = line.index(line.lstrip(' \t')[0])
            if last_index is None or last_index < first_index:
                last_index = first_index
                if idx == 0:
                    level += 1
            elif last_index > first_index:
                level -= 1
        else:
            level = start_level
        objs.append(ListItem(level=level, text=line.lstrip(' \t-')))
    return objs


def get_repo_name(full_name: str) -> str:
    return full_name[full_name.index('/') + 1:]


def get_reference_for_note(note: rnm.ReleaseNote) -> str:
    if note.is_current_repo:
        return note.reference_str
    return get_repo_name(
        note.source_component.name
    ) + note.reference.type.prefix + note.reference.identifier


def list_item_header_from_notes(
        line: str,
        note: rnm.ReleaseNote,
        category: Title,
        group: Title,
    ) -> ListItem:
    return ListItem(
        level=1,
        text=(
            f'`[{group.display}]` {line} by '
            f'{note.source_block.author or note.author} [{get_reference_for_note(note)}]'
        ),
    )


def list_item_from_note(
        message: str,
        note: rnm.ReleaseNote,
        group: Title,
) -> ListItem:
    # Replace newlines with two spaces _followed by_ newlines, as this is the proper way to do
    # a line-break in a list-item.
    # Also indent the next line, of course.
    message = message.replace('\n', '  \n  ')
    return ListItem(
        level=1,
        text=(
            f'`[{group.display}]` {message} by '
            f'{note.source_block.author or note.author} [{get_reference_for_note(note)}]'
        ),
    )


def render(notes: set[rnm.ReleaseNote]):
    objs = []

    # order by component
    components: dict[str, list[rnm.ReleaseNote]] = defaultdict(list)
    for note in notes:
        components[note.source_component.name].append(note)

    for component, notes in components.items():
        objs.append(Header(level=1, title=f'[{get_repo_name(component)}]'))

        # group by category
        cats: dict[Title, list[rnm.ReleaseNote]] = defaultdict(list)
        for note in notes:
            cat_title = categories_by_identifier.get(note.source_block.category)
            if not cat_title:
                logger.info(f"cannot find category '{note.source_block.category}'")
                continue
            cats[cat_title].append(note)

        # sort by category-priority, ascending. This will keep the order stable.
        for cat, notes in sorted(cats.items(), key=lambda tuple: tuple[0]):
            objs.append(Header(level=2, title=cat.display))

            # group by target group
            groups: dict[Title, list[rnm.ReleaseNote]] = defaultdict(list)
            for note in notes:
                group_title = target_groups_by_identifier.get(note.source_block.target_group)
                if not group_title:
                    logger.info(f"cannot find target group '{note.source_block.target_group}'")
                    continue
                groups[group_title].append(note)

            for group, notes in groups.items():
                for note in notes:
                    message = note.source_block.note_message
                    objs.append(list_item_from_note(message, note, group))
    return objs


def release_notes_for_ocm_resource(resource: ocm.Resource) -> str | None:
    if not resource.access:
        return None

    if resource.access.type is ocm.AccessType.OCI_REGISTRY:
        return f'- {resource.name}: `{resource.access.imageReference}`'

    return None


def release_note_for_ocm_component(component: ocm.Component) -> str | None:
    '''Create a markdown string containing information about the Resources included in the given
    Component.
    '''
    local_resources = [
        r
        for r in component.resources
        if r.relation is ocm.ResourceRelation.LOCAL
    ]

    component_release_notes = ''
    for resource_type in sorted({resource.type for resource in local_resources}):
        matching_resources = [r for r in local_resources if r.type == resource_type]
        resource_lines = {
            l for l in (
                release_notes_for_ocm_resource(r)
                for r in matching_resources
            ) if l is not None
        }

        if resource_type is ocm.ArtefactType.OCI_IMAGE:
            category_title = 'Docker Images'
        elif resource_type is ocm.ArtefactType.HELM_CHART:
            category_title = 'Helm Charts'
        else:
            category_title = str(resource_type)

        if resource_lines:
            category_markdown = (
                '## ' + category_title + '\n' + '\n'.join(sorted(resource_lines)) + '\n'
            )
            component_release_notes += category_markdown

    return component_release_notes



import subprocess

import ccc.github
import ci.util
import concourse.model.traits.version as version_trait


def has_version_conflict(
    target_tag: str,
    repository_name: str,
    repository_org: str,
    repository_hostname: str,
):
    github_cfg = ccc.github.github_cfg_for_repo_url(
        repo_url=ci.util.urljoin(repository_hostname, repository_org, repository_name),
    )
    github_api = ccc.github.github_api(github_cfg)

    target_tag = target_tag.removeprefix('refs/tags/')

    repository = github_api.repository(repository_org, repository_name)
    for tag in repository.tags():
        if tag.name == target_tag:
            return True

    return False


def read_version(
    version_interface: version_trait.VersionInterface,
    path: str,
):
    if version_interface is version_trait.VersionInterface.FILE:
        with open(path) as f:
            return f.read()
    elif version_interface is version_trait.VersionInterface.CALLBACK:
        res = subprocess.run(
            [path],
            capture_output=True,
            check=True,
            text=True,
        )

        version_str = res.stdout.strip()

        return version_str
    else:
        raise NotImplementedError


def write_version(
    version_interface: version_trait.VersionInterface,
    version_str: str,
    path: str,
):
    version_str = version_str.strip()

    if version_interface is version_trait.VersionInterface.FILE:
        with open(path, 'w') as f:
            f.write(version_str)
        return
    elif version_interface is version_trait.VersionInterface.CALLBACK:
        subprocess.run(
            [path],
            check=True,
            text=True,
            input=version_str,
        )
    else:
        raise NotImplementedError



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import urllib

from . import cluster_domain_from_kubernetes_config
from model.base import NamedModelElement


GARDENLINUX_CACHE_SUBDOMAIN_LABEL = 'snapshot-cache'


class GardenlinuxCacheConfig(NamedModelElement):
    '''Not intended to be instantiated by users of this module
    '''

    def namespace(self):
        return self.raw.get('namespace')

    def image_reference(self):
        return self.raw.get('image_reference')

    def external_url(self, cfg_factory):
        return self.ingress_host(cfg_factory)

    def image_tag(self):
        return self.raw.get('image_tag')

    def ingress_config(self):
        return self.raw.get('ingress_config')

    def ingress_host(self, cfg_factory):
        cluster_domain = cluster_domain_from_kubernetes_config(
            cfg_factory,
            self.kubernetes_config_name(),
        )
        return f'{self.subdomain_label()}.{cluster_domain}'

    def subdomain_label(self):
        return self.raw.get('subdomain_label', GARDENLINUX_CACHE_SUBDOMAIN_LABEL)

    def kubernetes_config_name(self):
        return self.raw.get('kubernetes_config')

    def volume_size(self) -> str:
        return self.raw.get('volume_size')

    def service_name(self):
        return self.raw.get('service_name')

    def service_port(self):
        return self.raw.get('service_port')

    def service_url(self):
        return urllib.parse.urlunparse((
            'http',
            f'{self.service_name()}.{self.namespace()}.svc.cluster.local:{self.service_port()}',
            '',
            '',
            '',
            '',
        ))

    def replicas(self):
        return self.raw.get('replicas')

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'image_reference',
            'image_tag',
            'ingress_config',
            'kubernetes_config',
            'namespace',
            'replicas',
            'service_name',
            'service_port',
            'volume_size',
        ]

    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'subdomain_label',
        ]



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import logging
import os

import ci.log


def configure_whd_logging(print_thread_id: bool=True):
    whd = logging.getLogger('whd')
    whd.setLevel(logging.DEBUG)
    whd_handler = logging.handlers.RotatingFileHandler(
        filename=os.path.join('/', 'tmp', 'whd_log'),
        backupCount=1,
        maxBytes=50 * 1024 * 1024,  # 50 MiB
    )
    whd_handler.setFormatter(
        ci.log.CCFormatter(
            fmt=ci.log.default_fmt_string(print_thread_id=print_thread_id)
        ),
    )
    whd.addHandler(whd_handler)



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import base64
import hashlib
import hmac
import typing

import reutil

import ctx
from model.base import (
    BasicCredentials,
    NamedModelElement,
)


def smtp_password_from_aws_access_key(
    aws_secret_access_key: str,
    aws_region: str,
):
    '''
    computes a password that be used for smtp-basic-auth for amazon's SES (simple email service)
    from given aws-secret-key and region

    stolen from https://docs.aws.amazon.com/ses/latest/dg/smtp-credentials.html
    '''
    # These values are required to calculate the credentials for AWS SES. Do not change them.
    date = "11111111"
    version = b'\x04'

    def sign(key, msg):
        return hmac.new(key, msg.encode('utf-8'), hashlib.sha256).digest()

    # do not change order or hard-coded values (see linked documentation above)
    signature = sign(f'AWS4{aws_secret_access_key}'.encode('utf-8'), date)
    signature = sign(signature, aws_region)
    signature = sign(signature, 'ses') # service
    signature = sign(signature, 'aws4_request') # terminal
    signature = sign(signature, 'SendRawEmail') # message
    signature_and_version = version + signature
    smtp_password = base64.b64encode(signature_and_version)

    return smtp_password.decode('utf-8')


class EmailCredentials(BasicCredentials):
    '''
    Not intended to be instantiated by users of this module
    '''
    pass


class AwsSesCredentials(EmailCredentials):
    def aws_config_name(self) -> str:
        return self.raw['aws_config_name']

    def username(self, cfg_factory=None) -> str:
        if not cfg_factory:
            cfg_factory = ctx.cfg_factory()
        return cfg_factory.aws(self.aws_config_name()).access_key_id()

    def passwd(self, cfg_factory=None) -> str:
        if not cfg_factory:
            cfg_factory = ctx.cfg_factory()

        aws_config = cfg_factory.aws(self.aws_config_name())

        aws_secret_access_key = aws_config.secret_access_key()
        aws_region = aws_config.region()

        return smtp_password_from_aws_access_key(
            aws_secret_access_key=aws_secret_access_key,
            aws_region=aws_region,
        )

    def _required_attributes(self):
        return ['aws_config_name']


class EmailConfig(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def smtp_host(self):
        return self.raw.get('host')

    def smtp_port(self):
        return self.raw.get('port')

    def use_tls(self):
        return self.raw.get('use_tls')

    def sender_name(self):
        return self.raw.get('sender_name')

    def credentials(self) -> EmailCredentials:
        if credentials := self.raw.get('credentials'):
            if 'aws_config_name' in credentials:
                return AwsSesCredentials(credentials)

        return EmailCredentials(credentials)

    def has_credentials(self):
        if self.raw.get('credentials'):
            return True
        return False

    def filter_recipients(self, recipients:typing.Iterable[str]):
        blacklist = self.raw.get('blacklist')
        if blacklist:
            email_filter = reutil.re_filter(exclude_regexes=blacklist)
            return {r for r in recipients if email_filter(r)}
        return recipients

    def _required_attributes(self):
        return ['host', 'port', 'credentials']

    def validate(self):
        super().validate()
        # ensure credentials are valid - validation implicitly happens in the constructor.
        if self.has_credentials():
            self.credentials()



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import logging

import slack
import slack.errors

import ci.log
import ci.util
import model.slack


logger = logging.getLogger(__name__)
ci.log.configure_default_logging()


class SlackHelper:
    def __init__(
            self,
            slack_cfg: model.slack.SlackConfig
    ):
        self.slack_cfg = slack_cfg

    def post_to_slack(
        self,
        channel: str,
        title: str,
        message: str,
        filetype: str='post'
    ):
        if not (api_token := self.slack_cfg.api_token()):
            raise RuntimeError("can't post to slack as there is no slack api token in config")

        logger.info(f'posting message {title=} to slack {channel=} using {self.slack_cfg.name()}')
        client = slack.WebClient(token=api_token)
        # We expect rather long messages, so we do not use incoming webhooks etc. to post
        # messages as those get truncated, see
        # https://api.slack.com/changelog/2018-04-truncating-really-long-messages
        # Instead we use the file upload mechanism so that this limit does not apply.
        # For contents of result see https://api.slack.com/methods/files.upload
        response = self._post_with_retry(
            client=client,
            retries=5,
            channels=channel,
            content=message,
            title=title,
            filetype=filetype,
        )
        if not response['ok']:
            raise RuntimeError(f"failed to post to slack channel '{channel}': {response['error']}")
        return response

    def _post_with_retry(self, client, retries=5, **kwargs):
        try:
            response = client.files_upload(**kwargs)
            return response
        except slack.errors.SlackApiError as sae:
            error_code = sae.response.get('error')
            error_status = sae.response.get('status')
            if retries < 1:
                raise sae # no retries left (or none requested)
            if error_code == 'markdown_conversion_failed_because_of_read_failed':
                logger.warning(f'received {error_code} - retrying {retries}')
                return self._post_with_retry(client=client, retries=retries-1, **kwargs)
            elif error_status in (503, 504): # Service Unavailable, Gateway Timeout
                logger.warning(
                    f"Slack responded with {error_status=} - retrying ({retries})"
                )
            else:
                raise sae # only retry for known sporadic err

    def delete_file(
        self,
        file_id: str,
    ):
        api_token = self.slack_cfg.api_token()
        if not api_token:
            raise RuntimeError("can't post to slack as there is no slack api token in config")
        logger.info(f"deleting file with id '{file_id}' from Slack")
        client = slack.WebClient(token=api_token)
        response = client.files_delete(
            file=file_id,
        )
        if not response['ok']:
            raise RuntimeError(f"failed to delete file with id {file_id}")
        return response


def post_to_slack(
    release_notes_markdown: str,
    component_name: str,
    slack_cfg_name: str,
    slack_channel: str,
    release_version: str,
    max_msg_size_bytes: int=20000,
):
    # XXX slack imposes a maximum msg size
    # https://api.slack.com/changelog/2018-04-truncating-really-long-messages#

    slack_cfg = ci.util.ctx().cfg_factory().slack(slack_cfg_name)
    slack_helper = SlackHelper(slack_cfg)

    idx = 0
    i = 0

    try:
        while True:
            title = f'[{component_name}:{release_version} released'

            # abort on last
            if idx + max_msg_size_bytes > len(release_notes_markdown):
                did_split = i > 0
                if did_split:
                    title += ' - final]'
                else:
                    title += ']'

                msg = release_notes_markdown[idx:]
                yield slack_helper.post_to_slack(channel=slack_channel, title=title, message=msg)
                break

            # post part
            title += f' - part {i} ]'
            msg = release_notes_markdown[idx: idx+max_msg_size_bytes]
            logger.info(f"Posting release-note '{title}'")
            yield slack_helper.post_to_slack(channel=slack_channel, title=title, message=msg)

            i += 1
            idx += max_msg_size_bytes

    except (RuntimeError, slack.errors.SlackApiError) as e:
        logger.warning(
            f'Unable to post release notes to Slack: {e}. Will dump generated notes next to '
            'enable manually posting them.'
        )
        print(release_notes_markdown)



import base64
import collections.abc
import dataclasses
import enum
import json
import operator
import os

import oci.util


class AuthType(enum.Enum):
    BASIC_AUTH = 'basic_auth'


class Privileges(enum.Enum):
    READONLY = 'readonly'
    READWRITE = 'readwrite'
    ADMIN = 'admin'

    def _asint(self, privileges):
        if privileges is self.READONLY:
            return 0
        elif privileges is self.READWRITE:
            return 1
        elif privileges is self.ADMIN:
            return 2
        elif privileges is None:
            return 4
        else:
            raise NotImplementedError(privileges)

    def __hash__(self):
        return self._asint(self).__hash__()

    def __lt__(self, other):
        o = self._asint(other)
        return self._asint(self).__lt__(o)

    def __le__(self, other):
        o = self._asint(other)
        return self._asint(self).__le__(o)

    def __eq__(self, other):
        o = self._asint(other)
        return self._asint(self).__eq__(o)

    def __ne__(self, other):
        o = self._asint(other)
        return self._asint(self).__ne__(o)

    def __gt__(self, other):
        o = self._asint(other)
        return self._asint(self).__gt__(o)

    def __ge__(self, other):
        o = self._asint(other)
        return self._asint(self).__ge__(o)


@dataclasses.dataclass(frozen=True)
class OciCredentials:
    pass


@dataclasses.dataclass(frozen=True)
class OciConfig:
    privileges: Privileges
    credentials: OciCredentials
    url_prefixes: collections.abc.Sequence[str] = dataclasses.field(default_factory=tuple)

    def valid_for(self, image_reference: str, privileges: Privileges=Privileges.READONLY):
        if privileges and privileges > self.privileges:
            return False

        if not self.url_prefixes:
            return True

        unmodified_ref = image_reference.lower()
        image_reference = oci.util.normalise_image_reference(image_reference=image_reference).lower()

        for prefix in self.url_prefixes:
            prefix = prefix.lower()

            if image_reference.startswith(oci.util.normalise_image_reference(prefix)):
                return True
            if image_reference.startswith(prefix.lower()):
                return True
            if unmodified_ref.startswith(prefix):
                return True

        return False


@dataclasses.dataclass(frozen=True)
class OciBasicAuthCredentials(OciCredentials):
    username: str
    password: str


# typehint-alias
image_reference = str
credentials_lookup = collections.abc.Callable[[image_reference, Privileges, bool], OciCredentials]


def mk_credentials_lookup(
    cfgs: OciCredentials | collections.abc.Sequence[OciCredentials],
) -> collections.abc.Callable[[image_reference, Privileges, bool], OciConfig]:
    '''
    returns a callable that can be queried for matching OciCredentials for requested
    privileges and image-references
    '''
    if isinstance(cfgs, OciConfig):
        cfgs = (cfgs,)

    def lookup_credentials(
        image_reference: str,
        privileges: Privileges=Privileges.READONLY,
        absent_ok: bool=False,
    ):
        valid_cfgs = sorted(
          (
            c for c in cfgs
            if c.valid_for(image_reference=image_reference, privileges=privileges)
          ),
          key=operator.attrgetter('privileges'),
        )

        if not valid_cfgs and absent_ok:
            return None

        if not valid_cfgs:
            raise ValueError(f'no valid cfg found: {image_reference=}, {privileges=}')

        # first element contains cfg with least required privileges
        return valid_cfgs[0].credentials

    return lookup_credentials


def docker_credentials_lookup(
    docker_cfg: str | None=None,
    absent_ok: bool=False,
) -> collections.abc.Callable[[image_reference, Privileges, bool], OciConfig]:
    '''
    returns a credentials-lookup backed by docker's auth-config. By design, docker's auth-config
    only allows configuring credentials per hostname. By default' docker-cfg is expected at
    `$HOME/.docker/config.json`. Location of docker-cfg can be customised via docker_cfg parameter.

    if no docker-cfg is found, raises RuntimeError, unless absent_ok is truthy, in which case the
    returned lookup will never return any credentials (which might still be useful for readonly
    operations that for many registries allow anonymous access).

    Note that docker does not offer to configure credentials by permissions, too (hence privileges
    parameter will be ignored)
    '''
    if not docker_cfg:
        docker_cfg = os.path.join(os.environ.get('HOME', ''), '.docker/config.json')

    if not os.path.isfile(docker_cfg):
        if not absent_ok:
            raise RuntimeError(f'not an existing file: {docker_cfg=}')

        def find_nothing_lookup(
            image_reference: str,
            privileges: Privileges=Privileges.READONLY,
            absent_ok: bool=False,
        ):
            if not absent_ok:
                raise ValueError(f'no auth-cfg found in {docker_cfg=} for {image_reference=}')
            return None

        return find_nothing_lookup

    def docker_auth_lookup(
        image_reference: str,
        privileges: Privileges=Privileges.READONLY,
        absent_ok: bool=False,
    ):
        # re-read docker-cfg to reflect fs-updates
        with open(docker_cfg) as f:
            docker_auth = json.load(f)
            auths = docker_auth.get('auths', None)

        if not auths:
            # docker-cfg might be empty - do not handle as an error; however, we can never serve
            # anything useful
            if not absent_ok:
                raise ValueError(f'no auth-cfg found in {docker_cfg=} for {image_reference=}')
            return None

        if image_reference.startswith('/'):
            # if it is a relative reference, we have not means to find appropriate cfg.
            if not absent_ok:
                raise ValueError(f'no auth-cfg found in {docker_cfg=} for {image_reference=}')
            return None

        # ignore ports - match cfg only by hostname
        image_netloc = image_reference.split('/')[0]
        image_host = image_netloc.split(':')[0]

        for netloc, auth_dict in auths.items():
            host = netloc.split(':')[0]
            if host == image_host:
                break
        else:
            if not absent_ok:
                raise ValueError(
                    f'no matching auth-cfg found in {docker_cfg=} for {image_reference=}'
                )
            return None # no matching cfg was found

        # we found a cfg
        # docker's auth-cfgs only have a single value `auth` (or so we hope / assume)
        auth = auth_dict.get('auth', None)
        if not auth:
            raise ValueError(f'did not find expected attr `auth` in {docker_cfg=} for {image_host=}')

        auth = base64.b64decode(auth).decode('utf-8')
        username, passwd = auth.split(':', 1)

        return OciBasicAuthCredentials(
            username=username,
            password=passwd,
        )

    return docker_auth_lookup



#!/usr/bin/env python3

# SPDX-FileCopyrightText: 2021 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0

import datetime
import enum
import json
import typing

import ocm


# adds the defined label to a list of labels. won't overwrite existing labels with the same key
def add_label(
    src_labels: typing.Sequence[ocm.Label],
    label: ocm.Label,
) -> typing.Sequence[ocm.Label]:
    label_exists = [src_label for src_label in src_labels if src_label.name == label.name]
    if label_exists:
        # label exists --> do not overwrite it
        return src_labels
    else:
        # label doesn't exist --> append it
        return src_labels + [
            label,
        ]


class EnumJSONEncoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, enum.Enum):
            return o.value
        elif isinstance(o, datetime.datetime):
            return o.isoformat()
        return super().default(o)



import base64
import copy
import json
import logging
import typing

import googleapiclient
import googleapiclient.errors

import ccc.gcp
import ccc.github
import cfg_mgmt
import cfg_mgmt.model as cmm
import ci.log
import ci.util
import model
import model.base
import model.container_registry
import model.gcp


ci.log.configure_default_logging()
logger = logging.getLogger(__name__)

GcpServiceAccount = model.container_registry.ContainerRegistryConfig | model.gcp.GcpServiceAccount


def _create_service_account_key(
    iam_client: googleapiclient.discovery.Resource,
    service_account_name: str,
) -> dict:
    '''
    Creates a key for a service account.
    '''

    key_request = iam_client.projects().serviceAccounts().keys().create(
        name=service_account_name,
        body={},
    )
    try:
        key = key_request.execute()
    except googleapiclient.errors.HttpError as e:
        logger.error('unable to create key')
        raise e

    logger.info('Created key: ' + key['name'])
    return json.loads(base64.b64decode(key['privateKeyData']))


def delete_service_account_key(
    iam_client: googleapiclient.discovery.Resource,
    service_account_key_name: str,
):
    try:
        iam_client.projects().serviceAccounts().keys().delete(
            name=service_account_key_name,
        ).execute()
        logger.info('Deleted key: ' + service_account_key_name)

    except googleapiclient.errors.HttpError as e:
        if not e.status_code == 404:
            raise

        logger.warning(f'{service_account_key_name=} not found, will remove entry from config queue')


def rotation_cfg_or_none(
    gcp_cfg: GcpServiceAccount,
    cfg_factory: model.ConfigFactory,
):
    if (rotation_cfg_reference := gcp_cfg.rotation_cfg()):
        return cfg_factory.resolve_cfg_element_reference(
            cfg_element_reference=rotation_cfg_reference,
        )

    return None


def rotate_cfg_element(
    cfg_element: GcpServiceAccount,
    cfg_factory: model.ConfigFactory,
) ->  typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]:
    rotation_cfg = rotation_cfg_or_none(
        gcp_cfg=cfg_element,
        cfg_factory=cfg_factory,
    ) or cfg_element

    logger.info(f'using {rotation_cfg.name()=} for rotation')

    client_email = cfg_element.client_email()

    iam_client = ccc.gcp.create_iam_client(
        cfg_element=rotation_cfg,
    )

    service_account_name = ccc.gcp.qualified_service_account_name(
        client_email,
    )

    old_key_id = cfg_element.private_key_id()
    old_key_id = ccc.gcp.qualified_service_account_key_name(
        service_account_name=client_email,
        key_name=old_key_id,
    )

    new_key = _create_service_account_key(
        iam_client=iam_client,
        service_account_name=service_account_name,
    )

    raw_cfg = copy.deepcopy(cfg_element.raw)

    if isinstance(cfg_element, model.container_registry.ContainerRegistryConfig):
        raw_cfg['password'] = json.dumps(new_key)
    elif isinstance(cfg_element, model.gcp.GcpServiceAccount):
        raw_cfg['service_account_key'] = new_key
    else:
        raise ValueError(cfg_element)

    updated_elem = type(cfg_element)(
        # checked for correct type already
        name=cfg_element.name(),
        raw_dict=raw_cfg,
        type_name=cfg_element._type_name,
    )

    secret_id = {'gcp_secret_key': old_key_id}

    def revert():
        delete_service_account_key(
            iam_client=iam_client,
            service_account_key_name=ccc.gcp.qualified_service_account_key_name(
                service_account_name=client_email,
                key_name=new_key['private_key_id'],
            )
        )

    return revert, secret_id, updated_elem


def delete_config_secret(
    cfg_element: GcpServiceAccount,
    cfg_queue_entry: cmm.CfgQueueEntry,
    cfg_factory: model.ConfigFactory,
) -> GcpServiceAccount | None:
    rotation_cfg = rotation_cfg_or_none(
        gcp_cfg=cfg_element,
        cfg_factory=cfg_factory,
    ) or cfg_element

    logger.info(f'using {rotation_cfg.name()=} for deletion')

    logger.info('deleting old gcr secret')
    iam_client = ccc.gcp.create_iam_client(
        cfg_element=rotation_cfg,
    )
    delete_service_account_key(
        iam_client=iam_client,
        service_account_key_name=cfg_queue_entry.secretId['gcp_secret_key'],
    )
    return None



import dataclasses
import dacite
import os

import ci.util


@dataclasses.dataclass
class SecretsServerTarget:
    type: str
    secrets_server_config: str
    secret_config: str
    kubernetes_config_name: str
    team_config: str


@dataclasses.dataclass
class KubernetesSecretTarget:
    type: str
    kubernetes_config_name: str
    secret_namespace: str
    secret_name: str
    secret_key: str


@dataclasses.dataclass
class ReplicationMapping:
    target: SecretsServerTarget | KubernetesSecretTarget
    cfg_set: str


@dataclasses.dataclass
class ReplicationTargetConfig:
    replication_mappings: list[ReplicationMapping]


def replication_config_from_cfg_dir(cfg_dir: str) -> ReplicationTargetConfig:
    replication_config = ci.util.parse_yaml_file(
        os.path.join(cfg_dir, 'config_replication_targets.yaml')
    )
    return dacite.from_dict(
        data_class=ReplicationTargetConfig,
        data=replication_config,
        config=dacite.Config(),
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import NamedModelElement


class AzureServicePrincipal(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def tenant_id(self) -> str:
        return self.raw['tenant_id']

    def client_id(self) -> str:
        return self.raw['client_id']

    def client_secret(self) -> str:
        return self.raw['client_secret']

    def object_id(self) -> str:
        return self.raw['object_id']

    def subscription_id(self) -> str:
        return self.raw['subscription_id']

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'client_id',
            'client_secret',
            'subscription_id',
            'tenant_id',
        ]


# XXX export previous name for - temporary - backwards compatibility
ServicePrincipal = AzureServicePrincipal


class AzureStorageAccountCfg(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def storage_account_name(self):
        return self.raw['storage_account_name']

    def access_key(self):
        return self.raw['access_key']

    def container_name(self):
        return self.raw['container_name']

    def container_name_sig(self):
        return self.raw['container_name_sig']

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'storage_account_name',
            'access_key',
            'container_name',
        ]


# XXX export previous name for - temporary - backwards compatibility
StorageAccountConfig = AzureStorageAccountCfg


class AzureSharedGalleryCfg(NamedModelElement):
    '''
    Not intended to be instantiated by users of this module
    '''

    def resource_group_name(self):
        return self.raw['resource_group_name']

    def gallery_name(self):
        return self.raw['gallery_name']

    def location(self):
        return self.raw['location']

    def published_name(self):
        return self.raw['published_name']

    def description(self):
        return self.raw['description']

    def eula(self):
        return self.raw['eula']

    def release_note_uri(self):
        return self.raw['release_note_uri']

    def identifier_publisher(self):
        return self.raw['identifier_publisher']

    def identifier_offer(self):
        return self.raw['identifier_offer']

    def identifier_sku(self):
        return self.raw['identifier_sku']

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'resource_group_name',
            'gallery_name',
            'location',
            'published_name',
        ]


# XXX export previous name for - temporary - backwards compatibility
SharedGalleryCfg = AzureSharedGalleryCfg



import oss2
from aliyunsdkcore.client import AcsClient
import ci.util


def _credentials(alicloud_cfg: str):
    if isinstance(alicloud_cfg, str):
        cfg_factory = ci.util.ctx().cfg_factory()
        alicloud_cfg = cfg_factory.alicloud(alicloud_cfg)
    return alicloud_cfg


def oss_auth(alicloud_cfg: str):
    cred = _credentials(alicloud_cfg)
    return oss2.Auth(cred.access_key_id(), cred.access_key_secret())


def acs_client(alicloud_cfg: str):
    cred = _credentials(alicloud_cfg)

    return AcsClient(
        ak=cred.access_key_id(),
        secret=cred.access_key_secret(),
        region_id=cred.region()
    )



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from html2text import html2text


def create_body(mail_template: str, replace_tokens: dict):
    for key, value in replace_tokens.items():
        mail_template = mail_template.replace(key, value)

    return mail_template


def create_mail(
        subject: str,
        sender: str,
        recipients: list[str],
        text: str,
        cc_recipients: list[str]=(),
        mimetype: str='text',
    ) -> MIMEMultipart:
    msg = MIMEMultipart('alternative')

    if mimetype == 'html':
        msg_html = MIMEText(
            text,
            mimetype,
            'utf-8',
        )
        plain_text = html2text(text)
        msg_plain = MIMEText(
            plain_text,
            'text',
            'utf-8',
        )
    elif mimetype == 'text':
        msg_html = None
        msg_plain = MIMEText(
            text,
            mimetype,
            'utf-8',
        )
    else:
        raise NotImplementedError()

    msg.attach(msg_plain)
    if msg_html:
        msg.attach(msg_html)

    msg['Subject'] = subject
    msg['From'] = sender
    msg['To'] = ','.join(recipients)
    msg['Cc'] = ','.join(cc_recipients)

    return msg



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from concourse.model.job import (
    JobVariant,
)
from concourse.model.step import (
    PipelineStep,
    PullRequestNotificationPolicy,
    StepNotificationPolicy,
)
from concourse.model.base import (
  TraitTransformer,
  ScriptType,
)

ENV_VAR_NAME = 'meta'
DIR_NAME = 'meta'
META_STEP_NAME = 'meta'


class MetaTraitTransformer(TraitTransformer):
    name = 'meta'

    def inject_steps(self):
        self.meta_step = PipelineStep(
            name='meta',
            raw_dict={},
            is_synthetic=True,
            notification_policy=StepNotificationPolicy.NO_NOTIFICATION,
            pull_request_notification_policy=PullRequestNotificationPolicy.NO_NOTIFICATION,
            injecting_trait_name=self.name,
            script_type=ScriptType.PYTHON3,
        )
        self.meta_step.add_output(name=DIR_NAME, variable_name=ENV_VAR_NAME)
        yield self.meta_step

    def process_pipeline_args(self, pipeline_args: JobVariant):
        # All steps depend on meta step and receive an input from it
        for step in pipeline_args.steps():
            if step == self.meta_step:
                continue
            step._add_dependency(self.meta_step)
            step.add_input(name=DIR_NAME, variable_name=ENV_VAR_NAME)
        if pipeline_args.has_trait('version'):
            # All steps depend on version. Remove ourself to avoid circular dependency
            version_step = pipeline_args.step('version')
            self.meta_step._remove_dependency(version_step)
            self.meta_step.remove_input('version_path')



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from concourse.model.base import (
    AttributeSpec,
    ModelBase,
)


OCI_IMAGE_CFG_ATTRIBUTES = (
    AttributeSpec.required(
        name='image_reference',
        type=str,
        doc='the OCI Image reference to use',
    ),
)


class OciImageCfg(ModelBase):
    @classmethod
    def _attribute_specs(cls):
        return OCI_IMAGE_CFG_ATTRIBUTES

    def image_reference(self):
        return self.raw['image_reference']


class FilterCfg(ModelBase):
    @classmethod
    def _attribute_specs(cls):
        return (
            AttributeSpec.optional(
                name='include_image_references',
                default=(),
                doc='''
                a list of regular expressions. If configured, only matching image references are
                processed. By default, all image references are considered.
                ''',
            ),
            AttributeSpec.optional(
                name='exclude_image_references',
                default=(),
                doc='''
                a list of regular expressions. If configured, matching image references are
                exempted from processing. Has precedence over include_image_references.
                By default, no image references are excluded.
                ''',
            ),
            AttributeSpec.optional(
                name='include_image_names',
                default=(),
                doc='''
                a list of regular expressions. If configured, only matching image names are
                processed. By default, all image references are considered.
                ''',
            ),
            AttributeSpec.optional(
                name='exclude_image_names',
                default=(),
                doc='''
                a list of regular expressions. If configured, matching image names are
                exempted from processing. Has precedence over include_image_references.
                By default, no image references are excluded.
                ''',
            ),
            AttributeSpec.optional(
                name='include_component_names',
                default=(),
                doc='''
                a list of regular expressions. If configured, only image references from components
                whose name matches are considered.
                ''',
            ),
            AttributeSpec.optional(
                name='exclude_component_names',
                default=(),
                doc='''
                a list of regular expressions. If configured, image references from components whose
                name matches are excluded from further processing. Has precedence over
                include_component_names.
                ''',
            ),
            AttributeSpec.optional(
                name='include_component_versions',
                default=[],
                doc='''
                a list of mappings of component-names to versions to be included. If
                configured, only these versions will be included for the given components.
                ''',
            ),
            AttributeSpec.optional(
                name='exclude_component_versions',
                default=[],
                doc='''
                a list of mappings of component-names to versions to be excluded. If
                configured, the given versions will be excluded for the given components.
                Takes precedence over include_component_versions.
                ''',
            ),
        )

    def include_image_references(self):
        return self.raw['include_image_references']

    def exclude_image_references(self):
        return self.raw['exclude_image_references']

    def include_image_names(self):
        return self.raw['include_image_names']

    def exclude_image_names(self):
        return self.raw['exclude_image_names']

    def include_component_names(self):
        return self.raw['include_component_names']

    def exclude_component_names(self):
        return self.raw['exclude_component_names']

    def include_component_versions(self):
        return self.raw['include_component_versions']

    def exclude_component_versions(self):
        return self.raw['exclude_component_versions']


class ImageFilterMixin(ModelBase):
    def filters(self):
        return FilterCfg(raw_dict=self.raw['filters'])


IMAGE_ATTRS = (
    AttributeSpec.optional(
        name='filters',
        default={
            'include_image_references': (),
            'exclude_image_references': (),
            'include_image_names': (),
            'exclude_image_names': (),
            'include_component_names': (),
            'exclude_component_names': (),
            'include_component_versions': [],
            'exclude_component_versions': [],
        },
        doc='optional filters to restrict container images to process',
        type=FilterCfg,
    ),
)



import dataclasses
import enum
import functools
import inspect
import sys

import dacite

import ocm
import dso.cvss


@dataclasses.dataclass(frozen=True)
class PathRegexes:
    include_paths: list[str] = dataclasses.field(default_factory=list)
    exclude_paths: list[str] = dataclasses.field(default_factory=list)


class ScanPolicy(enum.Enum):
    SCAN = 'scan'
    SKIP = 'skip'


@dataclasses.dataclass(frozen=True)
class LabelValue:
    pass


@dataclasses.dataclass(frozen=True)
class Label:
    name: str
    value: LabelValue


@dataclasses.dataclass(frozen=True)
class ScanningHint(LabelValue):
    policy: ScanPolicy
    path_config: PathRegexes | None
    comment: str | None


@dataclasses.dataclass(frozen=True)
class BinaryIdScanLabel(Label):
    name = 'cloud.gardener.cnudie/dso/scanning-hints/binary_id/v1'
    _alt_name = 'cloud.gardener.cnudie/dso/scanning-hints/binary/v1' # deprecated
    value: ScanningHint


@dataclasses.dataclass(frozen=True)
class SourceScanLabel(Label):
    name = 'cloud.gardener.cnudie/dso/scanning-hints/source_analysis/v1'
    value: ScanningHint


@dataclasses.dataclass(frozen=True)
class PackageVersionHint:
    name: str
    version: str


@dataclasses.dataclass(frozen=True)
class PackageVersionHintLabel(Label):
    name = 'cloud.gardener.cnudie/dso/scanning-hints/package-versions'
    value: tuple[PackageVersionHint]


@dataclasses.dataclass(frozen=True)
class SourceProjectLabel(Label):
    name = 'cloud.gardener.cnudie/dso/scanning-hints/checkmarx-project-name/v1'
    value: str


@dataclasses.dataclass(frozen=True)
class SourceScanHint(ScanningHint):
    pass


@dataclasses.dataclass(frozen=True)
class BinaryScanHint(ScanningHint):
    pass


@dataclasses.dataclass(frozen=True)
class SourceIdHint(ScanningHint):
    pass


@dataclasses.dataclass(frozen=True)
class CveCategorisationLabel(Label):
    name = 'gardener.cloud/cve-categorisation'
    value: dso.cvss.CveCategorisation


@dataclasses.dataclass(frozen=True)
class ExtraComponentReference:
    component_reference: ocm.ComponentIdentity
    purpose: tuple[str] | None


@dataclasses.dataclass(frozen=True)
class ExtraComponentReferencesLabel(Label):
    name = 'ocm.software/ocm-gear/extra-component-references'
    value: tuple[ExtraComponentReference, ...]


@functools.cache
def _label_to_type() -> dict[str, Label]:
    own_module = sys.modules[__name__]
    types = tuple(t for entry
        in inspect.getmembers(own_module, inspect.isclass)
        if (t := entry[1]) != Label and issubclass(t, Label)
    )

    label_names_to_types = {}
    for t in types:
        label_names_to_types[t.name] = t
        if (n := getattr(t, '_alt_name', None)):
            label_names_to_types[n] = t

    return label_names_to_types


def deserialise_label(
    label: ocm.Label | dict,
):
    if not (t := _label_to_type().get(label.name)):
        raise ValueError(f'unknown {label.name=}')

    if isinstance(label, ocm.Label):
        label = {
            'name': label.name,
            'value': label.value,
        }

    return dacite.from_dict(
        data_class=t,
        data=label,
        config=dacite.Config(
            cast=[tuple, enum.Enum],
        ),
    )



from model.base import (
    NamedModelElement,
)


class BtpServiceBinding(NamedModelElement):
    def credentials(self):
        '''
        service manager credentials (JSON) as retrieved from SAP BTP cockpit
        '''
        return self.raw['credentials']

    def instance_id(self):
        '''
        instance id of the service to create a service binding
        '''
        return self.raw['instance_id']

    def prefix(self):
        s = self.name()
        if s[-1] != '-':
            s += '-'
        return s

    def binding_id(self):
        '''
        binding id
        '''
        return self.raw['binding_id']

    def binding_name(self):
        '''
        binding name
        '''
        return self.raw['binding_name']

    def auth_service_binding(self):
        '''
        service binding used for authentication on service-manager
        '''
        return self.raw['auth_service_binding']

    def _required_attributes(self):
        return ['credentials', 'instance_id', 'binding_id', 'binding_name', 'auth_service_binding']



import json
import os
import uuid

import ci.util

import concourse.model.traits.meta
import concourse.paths


uuid_filename = 'job.uuid'
jobmetadata_filename = 'jobmetadata.json'


def get_out_dir():
    return os.path.join(
        ci.util.check_env('CC_ROOT_DIR'),
        concourse.model.traits.meta.DIR_NAME,
    )


def export_job_metadata(extra_attrs: dict={}):
    '''
    generates job metadata (currently only a UUID unambiguously identifying current build)
    and writes it into meta's output directory (hardcoded as contract)
    '''
    uuid_str = str(uuid.uuid4())
    metadata = {
        'uuid': uuid_str,
        **extra_attrs,
    }

    uuid_outfile = os.path.join(get_out_dir(), uuid_filename)
    with open(uuid_outfile, 'w') as f:
        f.write(uuid_str)

    jobmetadata_outfile = os.path.join(get_out_dir(), jobmetadata_filename)
    with open(jobmetadata_outfile, 'w') as f:
        json.dump(metadata, f)

    print(json.dumps(metadata, indent=2))



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


import typing

from concourse.model.base import (
  AttributeSpec,
  AttribSpecMixin,
  EnumWithDocumentation,
  EnumValueWithDocumentation,
  Trait,
  TraitTransformer,
  ModelBase,
  normalise_to_dict,
)
from concourse.model.job import (
    JobVariant,
)
from model.base import (
  NamedModelElement,
)


class NotificationTriggeringPolicy(EnumWithDocumentation):
    ONLY_FIRST = EnumValueWithDocumentation(
        value='only_first',
        doc='notify on first error only',
    )
    ALWAYS = EnumValueWithDocumentation(
        value='always',
        doc='notify on every error',
    )
    NEVER = EnumValueWithDocumentation(
        value='never',
        doc='notify never in case of errors',
    )


class NotificationRecipients(EnumWithDocumentation):
    EMAIL_ADDRESSES = EnumValueWithDocumentation(
        value='email_addresses',
        doc='''
                notifiy specific email addresses

                Example:

                .. code-block:: yaml

                    recipients:
                        - email_addresses:
                            - foo.bar@mycloud.com
                            - bar.buzz@mycloud.com
                ''',
    )

    COMMITTERS = EnumValueWithDocumentation(
        value='committers',
        doc='notify committers of the last commit',
    )

    COMPONENT_DIFF_OWNERS = EnumValueWithDocumentation(
        value='component_diff_owners',
        doc='notify the codeowners of a component. CODEOWNERS file must exist',
    )

    CODEOWNERS = EnumValueWithDocumentation(
        value='codeowners',
        doc='notify the codeowners of the repository. CODEOWNERS file must exist',
    )


class NotificationSlack(EnumWithDocumentation):
    CHANNEL_CFGS = EnumValueWithDocumentation(
        value='channel_cfgs',
        doc='''
                the slack channel configurations to use

                Example:

                .. code-block:: yaml

                    slack:
                      channel_cfgs:
                      - channel_names:
                        - 'my_slack_channel'
                        - 'my_other_slack_channel'
                        slack_cfg_name: 'my_slack_cfg_name'
                ''',
    )


NOTIFICATION_CFG_ATTRS = (
    AttributeSpec.optional(
        name='triggering_policy',
        default=NotificationTriggeringPolicy.ONLY_FIRST.value,
        doc='when to issue the configured notifications. Possible values see below',
        type=NotificationTriggeringPolicy,
    ),
    AttributeSpec.optional(
        name='email',
        default=True,
        doc='whether to send email notifications',
        type=bool,
    ),
    AttributeSpec.optional(
        name='inputs',
        default=['on_error_dir'],
        doc='configures the inputs that are made available to the notification',
        type=typing.List[str],
    ),
    AttributeSpec.optional(
        name='recipients',
        default=NotificationRecipients.COMMITTERS.value,
        doc='whom to notify. Possible values see blow.',
        type=NotificationRecipients,
    ),
    AttributeSpec.optional(
        name='cfg_callback',
        default=['on_error_dir'],
        doc='''
        an optional callback (relative to main repository root). Called as subprocess with
        an environment variables:

        - `REPO_ROOT`: absolute path to main repository
        - `NOTIFY_CFG_OUT`: absolute path to write notify.cfg to
        ''',
        type=typing.List[str],
    ),
    AttributeSpec.optional(
        name='slack',
        default=None,
        doc='send notification via Slack',
        type=NotificationSlack,
    )
)


class NotificationCfg(ModelBase):
    def __init__(self, raw_dict, *args, **kwargs):
        super().__init__(raw_dict=raw_dict, *args, **kwargs)
        self._apply_defaults(raw_dict=raw_dict)
        self.raw['recipients'] = normalise_to_dict(self.recipients())

    @classmethod
    def _attribute_specs(cls):
        return NOTIFICATION_CFG_ATTRS

    def triggering_policy(self):
        return NotificationTriggeringPolicy(self.raw['triggering_policy'])

    def should_send_email(self):
        return bool(self.raw.get('email'))

    def inputs(self):
        return self.raw.get('inputs')

    def recipients(self):
        return self.raw.get('recipients')

    def cfg_callback(self):
        return self.raw.get('cfg_callback')

    def slack(self):
        return self.raw.get('slack')


NOTIFICATION_CFG_SET_ATTRS = (
    AttributeSpec.optional(
        name='on_error',
        default={'triggering_policy': 'only_first'},
        doc='configures triggering policy in case a step fails with an error',
        type=NotificationCfg,
    ),
)


class NotificationCfgSet(NamedModelElement, AttribSpecMixin):
    def __init__(self, name, raw_dict, *args, **kwargs):
        super().__init__(name=name, raw_dict=raw_dict, *args, **kwargs)
        self._apply_defaults(raw_dict=raw_dict)

    @classmethod
    def _attribute_specs(cls):
        return NOTIFICATION_CFG_SET_ATTRS

    def _defaults_dict(self):
        return AttributeSpec.defaults_dict(self._attribute_specs())

    def on_error(self):
        return NotificationCfg(self.raw['on_error'])

    def _children(self):
        return (self.on_error(),)


ATTRIBUTES = (
    AttributeSpec.optional(
        name='default',
        default={
            'on_error': {
                'triggering_policy': 'only_first',
                'email': True
            }
        },
        doc='the default notification cfg (more may be defined)',
        type=NotificationCfgSet,
    ),
)


class NotificationsTrait(Trait):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @classmethod
    def _attribute_specs(cls):
        return ATTRIBUTES

    def _known_attributes(self):
        return self.raw.keys()

    def _children(self):
        return [NotificationCfgSet(name, raw_dict) for name, raw_dict in self.raw.items()]

    def notifications_cfg(self, cfg_name):
        return NotificationCfgSet(cfg_name, self.raw[cfg_name])

    def transformer(self):
        return NotificationsTraitTransformer(self)


class NotificationsTraitTransformer(TraitTransformer):
    name = 'notifications'

    def __init__(self, trait):
        self.trait = trait

    def process_pipeline_args(self, pipeline_args: JobVariant):
        # all steps depend from us and may consume our output
        for step in pipeline_args.steps():
            step._notifications_cfg = self.trait.notifications_cfg(step.notifications_cfg_name())



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from model.base import NamedModelElement


class DockerConfig(NamedModelElement):
    def username(self):
        return self.raw.get('username')

    def email_address(self):
        return self.raw.get('email_address')

    def password(self):
        return self.raw.get('password')

    def access_tokens(self):
        return self.raw.get('access_tokens')

    def _required_attributes(self):
        return (
            'username',
            'email_address',
            'password',
        )

    def _optional_attributes(self):
        return ('access_tokens',)



import unittest

import model
import model.github

from model.base import ModelValidationError
from concourse.factory import (
    RawPipelineDefinitionDescriptor as DefDescriptor,
    DefinitionFactory
)


class RawPipelineDefinitionDescriptorTest(unittest.TestCase):
    def test_basic_validation(self):
        # "OK case"
        DefDescriptor(name='a_name', base_definition=None, jobs={'foo': 42})

        # error cases
        with self.assertRaises(ValueError):
            DefDescriptor(name=None, base_definition=None, jobs={'foo': 42})
        with self.assertRaises(ModelValidationError):
            DefDescriptor(name='a_name', base_definition=None, jobs={})


class DefinitionFactoryTest(unittest.TestCase):
    def test_valid_validation(self):
        # "OK case"
        descriptor = DefDescriptor(name='a_name', base_definition=None, jobs={'foo': 42})
        DefinitionFactory(
            raw_definition_descriptor=descriptor,
            cfg_set=None,
        )

        with self.assertRaises(ValueError):
            DefinitionFactory(
                raw_definition_descriptor=None,
                cfg_set=None,
            )

    def test_inheritance(self):
        base_def = {'foo': 'bar', 123: 555}
        jobs = {
            'variant_a': {'foo': 42},
            'variant_b': {'xxx': 31}
        }
        descriptor = DefDescriptor(name='x_name', base_definition=base_def, jobs=jobs)
        factory = DefinitionFactory(
            raw_definition_descriptor=descriptor,
            cfg_set=None,
        )

        merged_variants = factory._create_variants_dict(descriptor)

        self.assertEqual(set(merged_variants.keys()), {'variant_a', 'variant_b'})

        variant_a = merged_variants['variant_a']
        variant_b = merged_variants['variant_b']

        # jobs may overwrite values
        self.assertEqual(variant_a, {'foo': 42, 123: 555})

        # jobs may add attributes
        self. assertEqual(variant_b, {'foo': 'bar', 123: 555, 'xxx': 31})

    def test_repository_triggers_logic(self):
        '''
        if a repository "triggers" a build, it will (apart from triggering the given build)
        receive a webhook configuration.
        By default, only the 'main' repository "triggers". Triggering can be explicitly
        configured using the 'trigger' attribute. Ensure that different trigger configurations
        for the same repository will be honoured properly for each variant.
        '''
        base_def = {'repo': {'name': 'main_repo', 'branch': 'dontcare', 'path': 'foo/bar'}}
        jobs = {
            'variant2':
            {
                'repo': {'name': 'main_repo', 'trigger': False},
                'repos':
                [
                    {'name': 'other_repo', 'branch': 'x_branch', 'path': 'x/path', 'trigger':True}
                ]
            },
            'variant1':
            {
                'repos': [{'name': 'other_repo', 'branch': 'x_branch', 'path': 'b/path'}]
            },
            'variant3':
            {
                'repo': {'name': 'main_repo', 'trigger': False}
            },
        }
        descriptor = DefDescriptor(name='foo', base_definition=base_def, jobs=jobs)

        cfg_set = model.ConfigurationSet('dummy','dummy', raw_dict={})
        cfg_set.github = unittest.mock.MagicMock(
            return_value=model.github.GithubConfig(
                    name='dontcare',
                    raw_dict={'available_protocols': ('https',)},
                    type_name='github',
                ),
            )
        factory = DefinitionFactory(
            raw_definition_descriptor=descriptor,
            cfg_set=cfg_set,
        )

        result = factory.create_pipeline_definition()
        variant = result.variant('variant1')

        main_repo = variant.repository('main_repo')
        other_repo = variant.repository('other_repo')

        # main_repo should "trigger" the job (and thus a webhook ought to be generated for it)
        self.assertTrue(main_repo.should_trigger())

        # non-main repos should not "trigger"
        self.assertFalse(other_repo.should_trigger())

        variant2 = result.variant('variant2')

        main_repo_v2 = variant2.repository('main_repo')
        other_repo_v2 = variant2.repository('other_repo')

        self.assertFalse(main_repo_v2.should_trigger())
        self.assertTrue(other_repo_v2.should_trigger())

        # ensure different trigger logic in resource_registry:
        # if any variant declares a repository to be triggering, this should be the result
        registry = result.resource_registry()
        main_repo_from_registry = registry.resource(main_repo.resource_identifier())

        self.assertTrue(main_repo_from_registry.should_trigger())



# SPDX-FileCopyrightText: 2024 SAP SE or an SAP affiliate company and Gardener contributors
#
# SPDX-License-Identifier: Apache-2.0


from . import cluster_domain_from_kubernetes_config
from model.base import NamedModelElement


TEKTON_INGRESS_SUBDOMAIN_LABEL = 'tekton-dashboard'


class TektonDashboardIngressConfig(NamedModelElement):
    '''Not intended to be instantiated by users of this module
    '''

    def namespace(self):
        return self.raw.get('namespace')

    def external_url(self, cfg_factory):
        return self.ingress_host(cfg_factory)

    def ingress_config(self):
        return self.raw.get('ingress_config')

    def ingress_host(self, cfg_factory):
        cluster_domain = cluster_domain_from_kubernetes_config(
            cfg_factory,
            self.kubernetes_config_name(),
        )
        return f'{self.subdomain_label()}.{cluster_domain}'

    def subdomain_label(self):
        return self.raw.get('subdomain_label', TEKTON_INGRESS_SUBDOMAIN_LABEL)

    def kubernetes_config_name(self):
        return self.raw.get('kubernetes_config')

    def service_name(self):
        return self.raw.get('service_name')

    def service_port(self):
        return self.raw.get('service_port')

    def oauth2_proxy_config_name(self):
        return self.raw.get('oauth2_proxy_config_name')

    def _required_attributes(self):
        yield from super()._required_attributes()
        yield from [
            'ingress_config',
            'kubernetes_config',
            'namespace',
            'service_name',
            'service_port',
        ]

    def _optional_attributes(self):
        yield from super()._optional_attributes()
        yield from [
            'subdomain_label',
        ]



import copy
import logging
import typing
import requests
import dacite
import datetime
import dateutil.parser
from dataclasses import dataclass

import cfg_mgmt
import cfg_mgmt.model as cmm
import ci.log
import ci.util
import model
import model.btp_service_binding
import model.container_registry


ci.log.configure_default_logging()
logger = logging.getLogger(__name__)


@dataclass
class ServiceBindingInfo:
    id: str
    name: str
    service_instance_id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime


@dataclass
class ServiceBindingDetailedInfo(ServiceBindingInfo):
    credentials: dict


def _find_next_service_binding_serial_number(
    bindings: list[ServiceBindingInfo],
    binding_name_prefix: str,
    instance_id: str,
) -> int:
    next_sn = 1
    for item in bindings:
        if item.name.startswith(binding_name_prefix) and item.service_instance_id == instance_id:
            try:
                n = int(item.name[len(binding_name_prefix):])
                if n >= next_sn:
                    next_sn = n + 1
            except ValueError:
                logger.warn(f'ignored {item.name}')
    return next_sn


class SBClient:
    def __init__(
        self,
        sm_url: str,
        access_token,
    ):
        self.sm_url = sm_url
        self.access_token = access_token

    def delete_service_binding(self, name: str, id: str):
        headers = {
            'Authorization': f'Bearer {self.access_token}',
        }
        url = f'{self.sm_url}/v1/service_bindings/{id}'
        resp = requests.delete(
            url,
            headers=headers,
            timeout=(4, 31),
        )
        if not resp.ok:
            msg = f'delete_service_binding failed: {resp.status_code} {resp.text}'
            logger.error(msg)
            raise requests.HTTPError(msg)
        logger.info(f'Deleted service binding {name} ({id})')

    def create_service_binding(self, instance_id: str, binding_name: str) \
        -> ServiceBindingDetailedInfo:
        url = f'{self.sm_url}/v1/service_bindings'
        data = {
            'name': binding_name,
            'service_instance_id': instance_id,
        }
        headers = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.access_token}',
        }
        resp = requests.post(
            url,
            json=data,
            headers=headers,
            timeout=(4, 31),
        )
        if not resp.ok:
            msg = f'create_service_binding failed: {resp.status_code} {resp.text}'
            logger.error(msg)
            raise requests.HTTPError(msg)
        result = resp.json()
        id = result['id']
        logger.info(f'Creating service binding {binding_name} ({id})')
        return self._response_json_to_sb_detailed_info(result)

    def get_service_binding(self, id: str) -> ServiceBindingDetailedInfo:
        url = f'{self.sm_url}/v1/service_bindings/{id}'
        headers = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.access_token}',
        }
        resp = requests.get(
            url,
            headers=headers,
            timeout=(4, 31),
        )
        if not resp.ok:
            msg = f'get_service_binding failed: {resp.status_code} {resp.text}'
            logger.error(msg)
            raise requests.HTTPError(msg)
        result = resp.json()
        return self._response_json_to_sb_detailed_info(result)

    def get_service_bindings(self) -> list[ServiceBindingInfo]:
        url = f'{self.sm_url}/v1/service_bindings'
        headers = {
            'Accept': 'application/json',
            'Authorization': f'Bearer {self.access_token}',
        }
        resp = requests.get(
            url,
            headers=headers,
            timeout=(4, 31),
        )
        if not resp.ok:
            msg = f'get_service_bindings failed: {resp.status_code} {resp.text}'
            logger.error(msg)
            raise requests.HTTPError(msg)
        bindings = []
        for item in resp.json()['items']:
            bindings.append(
                dacite.from_dict(
                    data=item,
                    data_class=ServiceBindingInfo,
                    config=dacite.config.Config(
                        type_hooks={
                            datetime.datetime: dateutil.parser.isoparse,
                        }
                    )
                )
            )
        return bindings

    def _response_json_to_sb_detailed_info(self, resp_json) -> ServiceBindingDetailedInfo:
        return dacite.from_dict(
            data=resp_json,
            data_class=ServiceBindingDetailedInfo,
            config=dacite.config.Config(
                type_hooks={
                    datetime.datetime: dateutil.parser.isoparse,
                }
            )
        )


def _get_oauth_token(credentials: dict) -> str:
    url = f'{credentials["url"]}/oauth/token'
    data = {
        'grant_type': 'client_credentials',
        'token_format': 'bearer',
        'client_id': credentials['clientid'],
        'client_secret': credentials['clientsecret'],
    }
    headers = {
        'Accept': 'application/json',
    }
    resp = requests.post(
        url,
        data=data,
        headers=headers,
        timeout=(4, 31),
    )
    if not resp.ok:
        msg = f'_get_oauth_token failed: {resp.status_code} {resp.reason}'
        logger.error(msg)
        raise requests.HTTPError(msg)
    result = resp.json()
    return result['access_token']


def _authenticate(
    cfg_element: model.btp_service_binding.BtpServiceBinding,
    cfg_factory: model.ConfigFactory,
) -> SBClient:
    auth = cfg_element.auth_service_binding()
    credentials = cfg_factory.btp_service_binding(auth).credentials()
    sm_url = credentials['sm_url']
    access_token = _get_oauth_token(credentials)
    return SBClient(sm_url, access_token)


def rotate_cfg_element(
    cfg_element: model.btp_service_binding.BtpServiceBinding,
    cfg_factory: model.ConfigFactory,
) -> typing.Tuple[cfg_mgmt.revert_function, dict, model.NamedModelElement]:
    old_binding_id = cfg_element.binding_id()
    old_binding_name = cfg_element.binding_name()
    client = _authenticate(cfg_element, cfg_factory)
    bindings = client.get_service_bindings()
    prefix = cfg_element.prefix()
    instance_id = cfg_element.instance_id()
    next_sn = _find_next_service_binding_serial_number(bindings, prefix, instance_id)
    binding_name = f'{prefix}{next_sn}'
    created_binding = client.create_service_binding(instance_id, binding_name)

    secret_id = {'binding_id': old_binding_id, 'binding_name': old_binding_name}
    raw_cfg = copy.deepcopy(cfg_element.raw)
    raw_cfg['credentials'] = created_binding.credentials
    raw_cfg['binding_id'] = created_binding.id
    raw_cfg['binding_name'] = binding_name
    updated_elem = model.btp_service_binding.BtpServiceBinding(
        name=cfg_element.name(), raw_dict=raw_cfg, type_name=cfg_element._type_name
    )

    def revert():
        client.delete_service_binding(binding_name, id)

    return revert, secret_id, updated_elem


def delete_config_secret(
    cfg_element: model.btp_service_binding.BtpServiceBinding,
    cfg_queue_entry: cmm.CfgQueueEntry,
    cfg_factory: model.ConfigFactory,
) -> model.btp_service_binding.BtpServiceBinding | None:
    logger.info('deleting old service binding')
    id = cfg_queue_entry.secretId.get('binding_id')
    if id:
        name = cfg_queue_entry.secretId['binding_name']
        client = _authenticate(cfg_element, cfg_factory)
        client.delete_service_binding(name, id)

    return None
